# Copyright 2019, Rasmus Sorensen <rasmusscholer@gmail.com>
"""

Module containing Todoist action commands for the todoist-action-cli.




"""
import operator
import sys
import builtins

import parsedatetime
from dateutil import tz
from todoist.models import Item

from actionista import binary_operators
# 'in' is a reserved keyword, so the equivalent command is `in_`:
setattr(binary_operators, 'in', binary_operators.in_)
from actionista.date_utils import ISO_8601_FMT, start_of_day, DATE_DAY_FMT, end_of_day
from actionista.date_utils import local_time_to_utc, get_rfc3339_datestr
from actionista.todoist.config import DEFAULT_TASK_PRINT_FMT, DEFAULT_TASK_SORT_KEYS, DEFAULT_TASK_SORT_ORDER
from actionista.todoist.config import get_config
from actionista.todoist.tasks_utils import get_task_value, get_recurring_tasks, is_recurring
from actionista.todoist.tasks_utils import inject_tasks_project_fields


CONFIG = get_config()


def print_tasks(
        tasks: list,
        print_fmt: str = DEFAULT_TASK_PRINT_FMT,
        header=None, sep: str = "\n",
        *,
        data_attr: str = "_custom_data",
        verbose: int = 0,
        config=None,
):
    """ Print tasks, using a python format string.

    Examples:
        `-print`
        `-print "{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M  } {content}"`
        `-print "{project_name:15} {content}" "Project name:   Task:`

    Args:
        tasks: List of tasks or task_data dicts to print.
        print_fmt: How to print each task.
            Note: You can use print_fmt="pprint" to just print all tasks using pprint.
        header: Print a header before printing the tasks.
        sep: How to separate each printed task. Default is just "\n".
        # Keyword only arguments:
        data_attr: The task attribute to get task data from.
            Original data from the Todoist Sync API is stored in Item.data,
            but I prefer to add derived data fields in a separate Item._custom_data,
            so that they don't get persisted when writing the cache to disk.
        verbose: The verbosity to print informational messages with during the filtering process.
        config: Optional configuration dict.

    Returns: List of tasks.

    Frequently-used print formats:
        "* {content}"
        "{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M  } {content}"
        "{project_name:15} {due_date_safe_iso}    {content}  ({checked})"
        "{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M}  {checked}  {content}"
        "{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M}  {checked_str}  {content}"

    Frequently-used headers:
        "Project:        Due_date:          Task:"
        "Project:        Due_date:          Done: Task:"
        "Project:        Due_date:        Done: Task:"
        "Project:        Due_date:        P:   Done: Task:"

    See `inject_tasks_date_fields()` for available date fields. In brief:
        due_date_utc
        due_date_utc_iso
        due_date_dt
        due_date_local_dt
        due_date_local_iso
        due_date_safe_dt   # These two '_safe_' fields are guaranteed to be present, even if the task
        due_date_safe_iso  # has no due_date, in which case we use the end of the century. (Note: local time!)
        # We also have the same above fields for `date_added` and `completed_date`.

    """
    if config is None:
        config = CONFIG
    if print_fmt is None:
        print_fmt = config.get('default_task_print_fmt', DEFAULT_TASK_PRINT_FMT) if config else DEFAULT_TASK_PRINT_FMT
    if verbose > -1:
        print(f"\n - Printing {len(tasks)} tasks",
              f"separated by {sep!r}, using print_fmt:\n{print_fmt!r}.\n" if verbose else "...\n",
              file=sys.stderr)
    if header:
        print(header)
    # Use `task._custom_data`, which is a copy of task.data with extra stuff added (if available).
    task_dicts = [getattr(task, data_attr, task.data) if isinstance(task, Item) else task for task in tasks]
    if print_fmt == 'repr' or print_fmt == 'pprint':
        import pprint
        pprint.pprint(task_dicts)
    else:
        print(sep.join(print_fmt.format(task=task, **task) for task in task_dicts))
    return tasks


def sort_tasks(tasks, keys=DEFAULT_TASK_SORT_KEYS, order=DEFAULT_TASK_SORT_ORDER,
               *, data_attr="_custom_data", verbose=0, config=None):
    """ Sort the list of tasks, by task attribute in ascending or descending order.

    Args:
        tasks: The tasks to sort (dicts or todoist.moddl.Item objects).
        keys: The keys to sort by. Should be a list or comma-separated string.
        order: The sort order, either ascending or descending.
        # Keyword only arguments:
        data_attr: Ues this attribute for task data. For instance, if the
        verbose: The verbosity to print informational messages with during the filtering process.

    Examples:

        Sort tasks by project_name, then priority, in descending order:
            -sort "project_name,priority" descending
            sort_tasks(tasks, keys="project_name,priority", order="descending")

    Frequently-used sortings:

        project_name,priority_str,item_order
        project_name,item_order
        due_date,priority,item_order

    """
    if config is None:
        config = CONFIG
    if keys is None:
        keys = config.get('default_task_sort_keys', DEFAULT_TASK_SORT_KEYS) if config else DEFAULT_TASK_SORT_KEYS
    if order is None:
        order = config.get('default_task_sort_order', DEFAULT_TASK_SORT_ORDER) if config else DEFAULT_TASK_SORT_ORDER
    if verbose > -1:
        print(f"\n - Sorting {len(tasks)} tasks by {keys!r} ({order}).", file=sys.stderr)
    if isinstance(keys, str):
        keys = keys.split(',')
    itemgetter = operator.itemgetter(*keys)
    if data_attr:
        def keyfunc(task):
            return itemgetter(getattr(task, data_attr, task.data))
    else:
        keyfunc = itemgetter
    tasks = sorted(tasks, key=keyfunc, reverse=(order == "descending"))
    return tasks


def filter_tasks(
        tasks,
        taskkey, op_name, value,
        missing="exclude", default=None,
        value_transform=None, negate=False,
        data_attr="_custom_data",
        *, verbose=0):
    """ Generic task filtering method based on comparison with a specific task attribute.

    CLI signature:
        $ todoist-action-cli -filter <taskkey> <operator> <value> <missing> <default> <transform> <negate>
    e.g.
        $ todoist-action-cli -filter project_id eq 2076120802 exclude none int

    Args:
        tasks: List of tasks (dicts or todoist.Item).
        taskkey: The attribute key to compare against, e.g. 'due_date_local_iso' or 'project'.
        op_name: Name of the binary operator to use when comparing the task attribute against the given value.
        value: The value to compare the task's attribute against.
        missing: How to deal with tasks with missing attributes, e.g. "include", "exclude", or default to a given value.
        default: Use this value if a task attribute is missing and missing="default".
        value_transform: Perform a given transformation of the input value, e.g. `int`, or `str`.
            value_transform can be the name of any function/class in the current namespace.
        negate: Can be used to negate (invert) the filter.
            Some operators already have an inverse operator, e.g. `eq` vs `ne`, `le` vs `gt`.
            But other operators do not have a simple inverse operator, e.g. `startswith`.
            So, if you want to remove/exclude tasks starting with 'Email', use:
                -filter content startswith Email exclude _ _ True
            Note: Negate applies to the transform, but not to tasks included/excluded due to missing value.
        data_attr: Instead of using task or task.data, use `getatr(task, data_attr)`.
            This is useful if you are setting custom task data on `task._custom_data` to keep them separate.
        verbose: The verbosity to print informational messages with during the filtering process.

    Returns:
        Filtered list of tasks passing the filter criteria (attribute matching value).

    The `missing` parameter controls what to do if `taskkey` is not present in a task:
        "raise" -> raise a KeyError.
        "include" -> include the task (task passes filter evaluation).
        "exclude" -> exclude the task (task fails filter evaluation).
        "default" -> Use a default value in leiu of the missing value.

    What if task[taskkey] is None?
        * Since we can't really use None for any comparison, it should be considered as missing,
            exactly the same as if the key was missing.

    """
    # First, check values and print helpful warnings about frequent pitfalls:
    if op_name == 'le' and 'date' in taskkey and value[-2:] != '59':
        print("\nWARNING: You are using the less-than-or-equal-to (`le`) operator with a data value, "
              "which can be tricky. Consider using the less-than (`lt`) operator instead. If you do use the "
              "less-than-or-equal-to (`le`) operator, make sure to specify full time in comparison.\n",
              file=sys.stderr)
    if taskkey == 'due_date_utc':
        print("\nNOTICE: You are using 'due_date_utc' as filter taskkey. This has the rather-unhelpful "
              "format: 'Mon 26 Mar 2018 21:59:59 +0000'.\n", file=sys.stderr)
    # We often use "_" as placeholeder on the command line, because we cannot enter e None value:
    if default == '_' or default == '__None__':
        default = None
    if value_transform == '_' or value_transform == '__None__':
        value_transform = None
    if isinstance(negate, str) and negate.lower() in ('false', '_', '__none__', '0'):
        # Enabling easier ways to specify `negate` on the command line.
        negate = False

    negate = bool(negate)
    if isinstance(value, str) and len(value) > 0 and value[0] == "!":
        # Using exclamation mark is an established way to make a negative filter,
        # e.g. `-label !habit` to filter-out tasks with the "habit" label.
        negate = not negate
        value = value[1:]

    op = getattr(binary_operators, op_name)
    # I'm currently negating explicitly in the all four 'missing' cases,
    # but I could also just have re-defined `op` as: `def op(a, b): return _op(a, b) != negate`

    if verbose > -1:
        print(f"\n - Filtering {len(tasks)} tasks with: {taskkey!r} {op_name} {value!r} "
              f"(missing={missing!r}, default={default!r}, value_transform={value_transform!r}, negate={negate!r}).",
              file=sys.stderr)

    if value_transform:
        if isinstance(value_transform, str):
            # It is either 'int' or a custom eval statement:
            if value_transform in builtins.__dict__:
                value_transform = getattr(builtins, value_transform)  # e.g. 'int'
                # print(f"Using `value_transform` {value_transform!r} from `builtins`...")
            elif value_transform in locals():
                value_transform = locals()[value_transform]
                # print(f"Using `value_transform` {value_transform!r} from `locals()`...")
            elif value_transform in globals():
                value_transform = globals()[value_transform]
                # print(f"Using `value_transform` {value_transform!r} from `globals()`...")
            else:
                if verbose > -1:
                    print(f"Creating filter value transform by `eval({value_transform!r})` ...", file=sys.stderr)
                t = eval(value_transform)
                if hasattr(t, '__call__'):
                    value_transform = t
                else:
                    # Was a one-off eval transform intended as: `value = eval('value*2')`.
                    value = t
                    value_transform = None  # Prevent further transformation

    # value_transform can be used to transform the filter/comparison value to e.g. an int or datetime object:
    if value_transform:
        # custom callable:
        value = value_transform(value)
        if default and missing == "default":  # Only try to transform task default value if we actually need it
            default = value_transform(default)

    # TODO: Remove this! Instead, use `get_task_value()` - and `value_transform(value)` to tranform comparison value.
    def get_value(task, default_=None):
        nonlocal value
        task = getattr(task, data_attr, task.data) if isinstance(task, Item) else task
        # return taskkey not in task or op(itemgetter(task), value)
        if 'due' in task and taskkey in ('due_date', 'due_date_utc'):
            # Support for v7.1 Sync API with separate 'due' dict attribute:
            # Although this may depend on how `inject_tasks_date_fields()` deals with v7.1 tasks.
            due_dict = task.get('due') or {}
            task_value = due_dict.get(taskkey.replace('due_', ''))  # items in the 'due' dict don't have 'due_' prefix
        else:
            task_value = task.get(taskkey, default_)
        # Coercing the comparison value may not be optimal for lists, e.g. `-filter label_names contains 'habit'`
        # Maybe only coerce for ints? (e.g. priority?)
        if task_value is not None and type(task_value) != type(value) and isinstance(task_value, int):
            # Note: We are converting the *comparison value*, not the task value:
            print("NOTICE: `type(task_value) != type(value)` - Coercing `value` to %s:" % type(task_value),
                  file=sys.stderr)
            value = type(task_value)(value)
        return task_value

    if missing == "raise":
        def filter_eval(task):
            task_value = get_value(task)
            if verbose > 0:
                print(f"\n - Evaluating: task[{taskkey!r}] = {task_value}  {op_name} ({op}) {value} "
                      f"for task {task['content']} (due: {get_task_value(task, 'due_date')}) ",
                      file=sys.stderr)
            if task_value is None:
                raise ValueError(f"Key {taskkey!r} not present (or None) in task {task['id']}: {task['content']}")
            return op(task_value, value) != negate  # This comparison with negate will negate if negate is True.
    elif missing == "include":
        def filter_eval(task):
            # return taskkey not in task or op(itemgetter(task), value)
            task_value = get_value(task)
            if verbose > 0:
                print(f"\n - Evaluating: task[{taskkey!r}] = {task_value}  {op_name} ({op}) {value} "
                      f"for task {task['content']} (due: {get_task_value(task, 'due_date')}) ",
                      file=sys.stderr)
            return task_value is None or (op(task_value, value) != negate)
    elif missing == "exclude":
        def filter_eval(task):
            # return taskkey in task and op(itemgetter(task), value)
            task_value = get_value(task)
            if verbose > 0:
                print(f"\n - Evaluating: task[{taskkey!r}] = {task_value}  {op_name} ({op}) {value} "
                      f"for task {task['content']} (due: {get_task_value(task, 'due_date')}) ",
                      file=sys.stderr)
            return task_value is not None and (op(task_value, value) != negate)
    elif missing == "default":
        def filter_eval(task):
            task_value = get_value(task, default)
            return op(task_value, value) != negate
        if default is None:
            print('\nWARNING: filter_tasks() called with missing="default" but no default value given (is None).\n',
                  file=sys.stderr)
    else:
        raise ValueError("Argument `missing` value %r not recognized." % (missing,))

    tasks = [task for task in tasks if filter_eval(task)]

    return tasks


def generic_args_filter_adaptor(tasks, taskkey, args, *, default_op='iglob', **kwargs):
    """ A generic adaptor for filter_tasks(), accepting custom *args list.

    Typical use case is to be able to process both of the following action requests
    with a single function:
        `-content RS123*` and `-content startswith RS123`.
    This adaptor function just uses the number of args given to determine if a
    binary operator was provided with the action request.

    Args:
        tasks: List of tasks, passed to filter_tasks.
        taskkey: The task attribute to filter on.
        default_op: The default binary operator to use, in the case the user did not specify one in args.
        *args: User-provided action args, e.g. ['RS123*"], or ['startswith', 'RS123']

    Returns:
        Filtered list of tasks.

    """
    assert len(args) >= 1
    if args[0] == 'not':
        negate = True
        args = args[1:]
    else:
        negate = False
    if len(args) == 1:
        # `-content RS123*`
        op_name, value, args = default_op, args[0], args[1:]
    else:
        # `-content startswith work`
        op_name, value, *args = args

    # print(f"generic_args_filter_adaptor: args={args}")  # debugging
    return filter_tasks(tasks, taskkey=taskkey, op_name=op_name, value=value, negate=negate, *args, **kwargs)


def special_is_filter(tasks, *args, **kwargs):
    """ Special -is filter for ad-hoc or frequently-used cases, e.g. `-is not checked`, etc.

    These are generally implemented on an as-needed basis.

    Args:
        tasks:
        *args:

    Returns:
        tasks: Filtered list of tasks.
    """
    if args[0] == 'not':
        negate = True
        args = args[1:]
    else:
        negate = False
    if args[0] == 'due' or args[0] == 'overdue':
        # Note: "-is due" is alias for "due today or overdue" which is equivalent to "due before tomorrow".
        if args[0:2] == ["due", "or", "overdue"] or args[0:2] == ["overdue", "or", "due"]:
            args[0:2] = ["due", "before", "tomorrow"]
        timefmt = ISO_8601_FMT
        taskkey = 'due_date_utc_iso'  # Switch to 'due_date_iso' (or 'due_date_dt')
        convert = None
        if args[0] == 'overdue':
            op_name = 'lt'
            convert = start_of_day
            when = "today"
        elif len(args) > 1:
            if args[1] in ('before', 'on', 'after'):
                when = args[2]
                if args[1] == 'before':
                    op_name = 'lt'
                    convert = start_of_day
                elif args[1] == 'on':
                    op_name = 'startswith'
                    timefmt = DATE_DAY_FMT
                else:  # args[2] == 'after':
                    op_name = 'gt'
                    convert = end_of_day
            else:
                # "-is due today", same as "-is due on today"
                when = args[1]
                op_name = 'startswith'
                timefmt = DATE_DAY_FMT
        else:
            # "-is due":
            when = "today"
            op_name = 'le'
            convert = end_of_day
        # Using dateparser.DateDataParser().get_date_data() instead of dateparser.parse() we get a 'period' indication:
        # date_data = dateparser.DateDataParser().get_date_data(when)
        # if date_data is None:
        #     raise ValueError("Could not parse due date %r" % (when,))
        # dt, accuracy = date_data['date_obj'], date_data['period']  # Max 'period' precision is 'day' :(
        # Using parsedatetime, since dateparser has a poor concept of accuracy:
        # parsedatetime also understands e.g. "in two days", etc.
        cal = parsedatetime.Calendar()
        # Note: cal.parse returns a time.struct_time, not datetime object,
        # use cal.parseDT() to get a datetime object. Or just dt = datetime.datetime(*dt[:6])
        dt, context = cal.parseDT(when, version=2)  # provide `version` to get a context obj.
        if not context.hasDate:
            raise ValueError("Could not parse due date %r" % (when,))
        if convert and not context.hasTime:
            # Only perform conversion, i.e. snap to start/end of day, when no time indication was provided:
            dt = convert(dt)
        utc_str = local_time_to_utc(dt, fmt=timefmt)
        # date_value_iso = dt.strftime(timefmt)
        # When we request tasks that are due, we don't want completed tasks, so remove these first:
        tasks = filter_tasks(tasks, taskkey="checked", op_name="eq", value=0, missing="include", **kwargs)
        # Then return tasks that are due as requested:
        # return filter_tasks(tasks, taskkey=taskkey, op_name=op_name, value=utc_str, negate=negate, **kwargs)
        # Update, 2019-Sep: Use local datetime object for comparison:
        # OBS: can't compare offset-naive and offset-aware datetimes - so make sure `dt` has tzinfo:
        dt = dt.astimezone(tz.tzlocal())
        # Discussion: Maybe use 'due_date_safe_dt', which where tasks with no due date is set to a distant future.
        return filter_tasks(tasks, taskkey='due_date_dt', op_name=op_name, value=dt, negate=negate,
                            missing='exclude', **kwargs)
    elif args[0] in ('checked', 'unchecked', 'complete', 'incomplete', 'completed', 'done'):
        # -is not checked
        if args[0][:2] in ('in', 'un'):
            checked_val = 0
        else:
            checked_val = 1
        taskkey = "checked"
        op_name = "eq"
        return filter_tasks(tasks, taskkey=taskkey, op_name=op_name, value=checked_val, negate=negate, **kwargs)
    elif args[0] == 'in':
        taskkey = "project_name"
        op_name = "eq"
        value = args[1]
        return filter_tasks(tasks, taskkey=taskkey, op_name=op_name, value=value, negate=negate, **kwargs)
    elif args[0] == 'recurring':
        """ `-is [not] recurring` filter.
        NOTE: The 'is_recurring' attribute is not officially exposed and "may be removed soon",
        c.f. https://github.com/Doist/todoist-api/issues/33.
        Until then, perhaps it is better to filter based on whether the due_string starts with the word "every". 
        """
        # taskkey = "is_recurring"
        # op_name = "eq"
        # value = 1
        # return filter_tasks(tasks, taskkey=taskkey, op_name=op_name, value=value, negate=negate)
        # -is not recurring : for recurring task : negate==True, startswith('every')==True => startswith == negate
        print(f"\n - Filtering {len(tasks)} tasks, excluding {'' if negate else 'non-'}recurring tasks...")
        return get_recurring_tasks(tasks, negate=negate)
    else:
        raise ValueError("`-is` parameter %r not recognized. (args = %r)" % (args[0], args))


def is_not_filter(tasks, *args, **kwargs):
    """ Convenience `-not` action, just an alias for `-is not`. Can be used as e.g. `-not recurring`."""
    args = ['not'] + list(args)
    return special_is_filter(tasks, *args, **kwargs)


def due_date_filter(tasks, *when, **kwargs):
    """ Special `-due [when]` filter. Is just an alias for `-is due [when]`. """
    args = ['due'] + list(when)  # Apparently *args is a tuple, not a list.
    return special_is_filter(tasks, *args, **kwargs)


def content_filter(tasks, *args, **kwargs):
    """ Convenience adaptor to filter tasks based on the 'content' attribute (default op_name 'iglob'). """
    return generic_args_filter_adaptor(tasks=tasks, taskkey='content', args=args, **kwargs)


def content_contains_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="contains". """
    return filter_tasks(tasks, taskkey="content", op_name="contains", value=value, *args, **kwargs)


def content_startswith_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="startswith". """
    return filter_tasks(tasks, taskkey="content", op_name="startswith", value=value, *args, **kwargs)


def content_endswith_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="endswith"."""
    return filter_tasks(tasks, taskkey="content", op_name="endswith", value=value, *args, **kwargs)


def content_glob_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="glob". """
    return filter_tasks(tasks, taskkey="content", op_name="glob", value=value, *args, **kwargs)


def content_iglob_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="iglob". """
    return filter_tasks(tasks, taskkey="content", op_name="iglob", value=value, *args, **kwargs)


def content_eq_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="eq". """
    return filter_tasks(tasks, taskkey="content", op_name="eq", value=value, *args, **kwargs)


def content_ieq_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="ieq". """
    return filter_tasks(tasks, taskkey="content", op_name="ieq", value=value, *args, **kwargs)


def project_filter(tasks, *args, **kwargs):
    """ Convenience adaptor for filter action using taskkey="project_name" (default op_name "iglob"). """
    return generic_args_filter_adaptor(tasks=tasks, taskkey='project_name', args=args, **kwargs)


def project_iglob_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="content", op_name="iglob". """
    return filter_tasks(tasks, taskkey="project_name", op_name="iglob", value=value, *args, **kwargs)


def label_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="label_names", op_name="iin".
    OBS: The syntax is `-filter <taskkey> <op> <value>`,
    so we should be using `contains` as the op name.
    """
    return filter_tasks(tasks, taskkey="label_names", op_name="icontains", value=value, *args, **kwargs)


def priority_filter(tasks, *args, **kwargs):
    """ Convenience adaptor for filter action using taskkey="priority" (default op_name "eq"). """
    return generic_args_filter_adaptor(
        tasks=tasks, taskkey='priority', args=args, default_op='eq', value_transform=int, **kwargs)


def priority_ge_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="priority", op_name="ge". """
    value = int(value)
    return filter_tasks(tasks, taskkey="priority", op_name="ge", value=value, *args, **kwargs)


def priority_eq_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="priority", op_name="eq". """
    value = int(value)
    return filter_tasks(tasks, taskkey="priority", op_name="eq", value=value, *args, **kwargs)


def priority_str_filter(tasks, *args, **kwargs):
    """ Convenience adaptor for filter action using taskkey="priority_str" (default op_name "eq"). """
    # return filter_tasks(tasks, taskkey="priority_str", op_name="eq", value=value, *args)
    return generic_args_filter_adaptor(
        tasks=tasks, taskkey='priority_str', args=args, default_op='eq', **kwargs
    )


def priority_str_eq_filter(tasks, value, *args, **kwargs):
    """ Convenience filter action using taskkey="priority_str", op_name="eq". """
    return filter_tasks(tasks, taskkey="priority_str", op_name="eq", value=value, *args, **kwargs)


def p1_filter(tasks, *args, **kwargs):
    """ Filter tasks including only tasks with priority 'p1'. """
    return priority_str_eq_filter(tasks, value="p1", *args, **kwargs)


def p2_filter(tasks, *args, **kwargs):
    """ Filter tasks including only tasks with priority 'p2'. """
    return priority_str_eq_filter(tasks, value="p2", *args, **kwargs)


def p3_filter(tasks, *args, **kwargs):
    """ Filter tasks including only tasks with priority 'p3'. """
    return priority_str_eq_filter(tasks, value="p3", *args, **kwargs)


def p4_filter(tasks, *args, **kwargs):
    """ Filter tasks including only tasks with priority 'p3'. """
    return priority_str_eq_filter(tasks, value="p4", *args, **kwargs)


def reschedule_tasks(tasks, new_date, *, verbose=0):
    """ Reschedule tasks to a new date/time.

    Args:
        tasks: List of tasks.
        new_date: The new due_date string to send.
        verbose: Adjust the verbosity to increase or decrease the amount of information
            printed during function run.

    Returns:
        List of tasks.

    This command treats recurring and non-recurring tasks differently:

        * Regular, non-recurring tasks are rescheduled by updating due.string = new_date.
        * Recurring tasks are rescheduled by updating due.date = rfc3339(new_date),
            but keeping due.string as-is, so that the recurrence schedule won't change.

    If you would like to force one specific behavior, regardless of whether the task
    is recurring or not, please use one of the following functions:

        reschedule_tasks_due_date()  - this will always update due.date.
        reschedule_tasks_by_due_string() - this will always update due.string.

    If you would like to reschedule to a due date with a *fixed timezone*, please use:

        reschedule_tasks_fixed_timezone() - this will update due.date and due.timezone.

    Example: Reschedule tasks for tomorrow:

        >>> reschedule_tasks(tasks, 'tomorrow')

    CLI Example: Reschedule overdue tasks for tomorrow

        $ todoist-action-cli -sync -due before today -reschedule tomorrow

    API refs:
    * https://developer.todoist.com/sync/v8/#create-or-update-due-dates

    """
    if verbose > -1:
        print("\n - Rescheduling %s tasks for %r..." % (len(tasks), new_date), file=sys.stderr)

    date_rfc3339 = get_rfc3339_datestr(new_date)
    for task in tasks:
        if is_recurring(task):
            # Adjust "due.date", but keep "due.string" as-is:
            # OBS: It is probably still better to just use `-close` for repeating tasks:
            params = dict(due={
                "date": date_rfc3339,
                "string": get_task_value(task, 'due_string_safe'),
                "is_recurring": True,
            })
            task.update(**params)
            if verbose > 0:
                print(f" - Rescheduling task {task['content']} ({get_task_value(task, 'due_string_safe')}) "
                      f"with params: {params} ",
                      file=sys.stderr)
        else:
            # Adjust due.string, let server parse it:
            task.update(due={"string": new_date})
            if verbose > 0:
                print(f" - Rescheduling task {task['content']} for due.string={new_date}", file=sys.stderr)

    if verbose > -1:
        print("\n --> OBS: Remember to use `-commit` to push the changes (not `-sync`)! <--\n\n", file=sys.stderr)

    return tasks


def reschedule_tasks_due_date(tasks: list, due_date: str, *, verbose=0):
    """ Reschedule a recurring task by updating due.date but leaving due.string as-is. """
    date_rfc3339 = get_rfc3339_datestr(due_date)
    if verbose > -1:
        print(f"\n - Rescheduling {len(tasks)} tasks for '{due_date}' ({date_rfc3339}) ...",
              file=sys.stderr)
        print(" - Remember to use `-commit` to push the changes (not `-sync`)!\n\n", file=sys.stderr)
    for task in tasks:
        task.update(due={"date": date_rfc3339})
        # OBS: Other task date fields aren't updated until you've committed the changes!
    return tasks


def reschedule_tasks_by_due_string(tasks: list, due_string: str, *, check_recurring=True, verbose: int = 0):
    """ Reschedule tasks using due.string, letting the server parse the string. """
    if verbose > -1:
        print(f"\n - Rescheduling {len(tasks)} tasks for '{due_string}' ...", file=sys.stderr)
        print(" - Remember to use `-commit` to push the changes (not `-sync`)!\n\n", file=sys.stderr)
    if check_recurring is True:
        recurring_tasks = get_recurring_tasks(tasks)
        if len(recurring_tasks) > 0:
            print("\nWARNING: One or more of the tasks being rescheduled is currently a recurring task:")
            print_tasks(recurring_tasks)
            print("\n")
    for task in tasks:
        task.update(due={"string": due_string})
        # OBS: Other task date fields aren't updated until you've committed the changes!
    return tasks


def reschedule_tasks_fixed_timezone(tasks: list, due_string: str, timezone: str, *, verbose: int = 0):
    """ Reschedule tasks using due.date with fixed timezone.
    Again, the easiest solution is to just let the Todoist server parse the user input string.
    """
    if verbose > -1:
        print(f"\n - Rescheduling {len(tasks)} tasks for '{due_string}' "
              f"with fixed timezone '{timezone}' ...", file=sys.stderr)
        print(" - Remember to use `-commit` to push the changes (not `-sync`)!\n\n", file=sys.stderr)
    for task in tasks:
        task.update(due={"string": due_string, "timezone": timezone})
        # OBS: Other task date fields aren't updated until you've committed the changes!
    return tasks


def update_tasks(tasks, *, verbose=0, **kwargs):
    """ Generic task updater. (NOT FUNCTIONAL)

    """
    if verbose > -1:
        print("Updating tasks using kwargs:", kwargs, file=sys.stderr)
    for task in tasks:
        task.update(**kwargs)
    return tasks


def rename_tasks(tasks, new_content, *, verbose=0):
    """ Rename selected tasks, using API method 'item_update'. """
    if verbose > -1:
        print(f'\n - Renaming {len(tasks)} selected task(s) to "{new_content}" ...')
    for task in tasks:
        if verbose > -1:
            print(f' - Renaming task: "{task.data["content"]}" --> "{new_content}"')
        task.update(content=new_content)
    return tasks


def mark_tasks_completed(tasks, method='close', *, verbose=0):
    """ Mark tasks as completed using method='close'.

    Note: The Todoist v7 Sync API has two command types for completing tasks:
        'item_complete' - used by ?
        'item_update_date_complete' - used to mark recurring task completion.
        'item_close' - does exactly what official clients do when you close a task:
            regular task is completed and moved to history,
            subtasks are checked (marked as done, but not moved to history),
            recurring task is moved forward (due date is updated).
            Aka: "done".

    See: https://developer.todoist.com/sync/v7/#complete-items

    Args:
        tasks:  List of tasks.
        method: The method used to close the task.
            There are several meanings of marking a task as completed, especially for recurring tasks.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after closing them).

    See also:
        'item_uncomplete' - Mark task as uncomplete (re-open it).

    """
    if verbose > 0:
        print(f"\nMarking tasks as complete using method {method!r}...", file=sys.stderr)
        if method in ('close', 'item_close'):
            print(f"\nOBS: Consider using `-close` command directly instead ...", file=sys.stderr)
        print(" --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)

    for task in tasks:
        if method in ('close', 'item_close'):
            task.close()
        elif method in ('complete', 'item_complete'):
            task.complete()
        elif method in ('item_update_date_complete', 'complete_recurring'):
            raise NotImplementedError(
                f"Using method {method!r} is not implemented. "
                f"If using the CLI, please use the `-close` action instead. "
                f"If calling from Python, please use either close_tasks() function, or "
                f"`task.update_date_complete()`. ")
        else:
            raise ValueError(f"Value for method = {method!r} not recognized!")
    return tasks


def close_tasks(tasks, *, verbose=0):
    """ Mark tasks as completed using method='close'.

    See mark_tasks_completed for more info on the different API methods to "complete" a task.

    See: https://developer.todoist.com/sync/v8/#close-item

    Args:
        tasks:  List of tasks.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after closing them).

    See also:
        'item_uncomplete' - Mark task as uncomplete (re-open it).

    """
    if verbose > -1:
        print(f"\nClosing tasks (using API method 'item_close') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)
    for task in tasks:
        task.close()
    return tasks


def complete_and_update_date_for_recurring_tasks(tasks, new_date=None, due_string=None, *, verbose=0):
    """ Mark tasks as completed using method='item_update_date_complete'.

    See mark_tasks_completed for more info on the different API methods to "complete" a task.

    See: https://developer.todoist.com/sync/v8/#close-item

    Args:
        tasks:  List of tasks.
        new_date: The new/next due/occurrence date for the recurring task,
            e.g. `new_date="2019-09-
        due_string: Change the "due string" that specifies when the task occurs,
            e.g. `due_string="every monday 5 pm"`.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after completing/updating them).

    See also:
        'item_update_date_complete' - Mark task as uncomplete (re-open it).

    """
    if verbose > 0:
        print(f"\nCompleting recurring tasks and moving the due date to {new_date if new_date else 'next occurrence'} "
              f"(using API method 'item_update_date_complete') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)
    print("NOTICE: todoist.models.Item.update_date_complete() is currently broken in "
          "todoist-python package version 8.0.0.")
    for task in tasks:
        task.update_date_complete(new_date, due_string)
    return tasks


def uncomplete_tasks(tasks, *, verbose=0):
    """ Re-open tasks (uncomplete tasks) using API method 'item_uncomplete'.

    See: https://developer.todoist.com/sync/v8/#uncomplete-item

    Args:
        tasks:  List of tasks.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after reopening them).

    """
    if verbose > 0:
        print(f"\nRe-opening tasks (using API method 'item_uncomplete') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)
    for task in tasks:
        task.uncomplete()
    return tasks


def archive_tasks(tasks, *, verbose=0):
    """ Archive tasks using API method 'item_archive'.

    See: https://developer.todoist.com/sync/v8/?shell#archive-item

    Args:
        tasks:  List of tasks.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after archiving them).

    """
    if verbose > 0:
        print(f"\nArchiving tasks (using API method 'item_archive') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)
    for task in tasks:
        task.archive()
    return tasks


def delete_tasks(tasks, *, verbose=0):
    """ Delete tasks using API method 'item_delete'.

    See: https://developer.todoist.com/sync/v8/?shell#delete-item

    Args:
        tasks:  List of tasks.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after deleting them).

    """
    if verbose > 0:
        print(f"\nArchiving tasks (using API method 'item_delete') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)
    for task in tasks:
        task.delete()
    return tasks


def add_task(tasks, api, task_content, *, project=None, due=None, priority=None, labels=None,
             auto_reminder=True, auto_parse_labels=True, verbose=0):
    """ Add a new task API method 'item_add'.

    See: https://developer.todoist.com/sync/v8/?shell#add-an-item

    Args:
        tasks:  List of tasks.
        api: tooist.api.TodoistAPI object (so we can get projects, labels, etc.)
        due: Due date (str)
        project: Project name (str)
        labels: List of task labels (names)
        priority: Task priority, either number [1-4] or str ["p4", "p3", 'p2", "p1"].
        auto_reminder: Automatically add a default reminder to the task, if task is due at a specific time of day.
        auto_parse_labels: Automatically extract "@label" strings from the task content.
        verbose: Increase or decrease the verbosity of the information printed during function run.

    Returns:
        tasks:  List of tasks (after deleting them).

    """
    if verbose > -1:
        print(f"\nAdding new task (using API method 'item_add') ...", file=sys.stderr)
        print("\n --> Remember to `-commit` the changes to the server! <--", file=sys.stderr)

    params = {'auto_reminder': auto_reminder, 'auto_parse_labels': auto_parse_labels}

    if due:
        params['due'] = {"string": due}

    if project:
        if isinstance(project, str):
            # Assume project-name:
            # We need to find project_id from project_name:
            # If projects is e.g. a list of projects, create a dict mapping project_id to project:
            project_name = project
            # projects_by_name = {p['name']: p for p in api.projects.all()}
            # project_id = projects_by_name[project]['id']
            project_id = next(iter(p['id'] for p in api.projects.all() if p['name'] == project_name))
        else:
            # Project should be a project_id:
            assert isinstance(project, int)
            project_id = project
        params['project_id'] = project_id

    if labels:
        if isinstance(labels, str):
            labels = [label.strip() for label in labels.split(",")]
        if any(isinstance(label, str) for label in labels):
            # Assume label-name: We need to find label_id from label_name:
            labels_by_name = {label['name']: label for label in api.labels.all()}
            labels = [label if isinstance(label, int) else labels_by_name[label]['id'] for label in labels]
        params['labels'] = labels

    if priority:
        if isinstance(priority, str):
            # Priority in the form of "p1" to "p4".
            priority_str = priority
            priority_str_map = dict(p1=4, p2=3, p3=2, p4=1)
            try:
                priority = priority_str_map[priority]
            except KeyError:
                raise ValueError("Argument `priority` must be one of 'p1', 'p2', 'p3', or 'p4' (if str).")
        else:
            # Integer between 1 and 4:
            if not isinstance(priority, int):
                raise TypeError("Argument `priority` must be str or int.")
            if priority < 1 or 4 < priority:
                raise ValueError("Argument `priority` must be between 1 and 4 (if int).")
        params['priority'] = priority

    new_task = api.items.add(task_content, **params)
    # tasks.append(new_task)  # This should not be needed, because api.items.add() updates the `tasks` list.

    return tasks


def fetch_completed_tasks(tasks, *, verbose=0):
    """ This will replace `tasks` with a list of completed tasks dicts. May not work nicely. Only for playing around.

    You should probably use the old CLI instead:
        $ todoist print-completed-today --print-fmt "* {title}"
    """
    if verbose > -1:
        print(f"Discarting the current {len(tasks)} tasks, and fetching completed tasks instead...")
    from actionista.todoist.adhoc_cli import completed_get_all
    # This will use a different `api` object instance to fetch completed tasks:
    tasks, projects = completed_get_all()
    inject_tasks_project_fields(tasks, projects)
    return tasks


# Defined ACTIONS dict AFTER we define the functions.
# OBS: ALL action functions MUST return tasks, never None.

ACTIONS = {
    'print': print_tasks,
    'sort': sort_tasks,
    'filter': filter_tasks,
    'has': filter_tasks,  # Undocumented alias, for now.
    'is': special_is_filter,  # Special cases, e.g. "-is incomplete" or "-is not overdue".
    'not': is_not_filter,
    'due': due_date_filter,
    # contains, startswith, glob/iglob, eq/ieq are all trivial derivatives of filter:
    # But they are special in that we use the binary operator name as the action name,
    # and assumes we want to filter the tasks, using 'content' as the task key/attribute.
    'contains': content_contains_filter,
    'startswith': content_startswith_filter,
    'endswith': content_endswith_filter,
    'glob': content_glob_filter,
    'iglob': content_iglob_filter,
    'eq': content_eq_filter,
    'ieq': content_ieq_filter,
    # Convenience actions where action name specifies the task attribute to filter on:
    'content': content_filter,  # `-content endswith "sugar".
    'name': content_filter,  # Alias for content_filter.
    'project': project_filter,
    # Label filter:
    'label': label_filter,  # Alias for `-filter label_names icontains <value>`
    # Priority filter (where priority=4 is higher than priority=1)
    'priority': priority_filter,
    'priority-eq': priority_eq_filter,
    'priority-ge': priority_ge_filter,
    'priority-str': priority_str_filter,
    'priority-str-eq': priority_str_eq_filter,
    # priority_str filters (where "p1" has higher priority than "p4")
    'p1': p1_filter,
    'p2': p2_filter,
    'p3': p3_filter,
    'p4': p4_filter,
    # Reschedule task actions:
    'reschedule': reschedule_tasks,
    'reschedule-due-date': reschedule_tasks_due_date,
    'reschedule-by-string': reschedule_tasks_by_due_string,
    'reschedule-fixed-timezone': reschedule_tasks_fixed_timezone,
    # Rename task:
    'rename': rename_tasks,
    # Complete task actions:
    'mark-completed': mark_tasks_completed,
    # 'mark-as-done': mark_tasks_completed,  # Deprecated.
    'close': close_tasks,
    'reopen': uncomplete_tasks,  # alias for uncomplete
    'uncomplete': uncomplete_tasks,
    'archive': archive_tasks,
    'complete_and_update': complete_and_update_date_for_recurring_tasks,
    # 'fetch-completed': fetch_completed_tasks,  # WARNING: Not sure how this works, but probably doesn't work well.
    # The following actions are overwritten when the api object is created inside the action_cli() function:
    'verbose': None, 'v': None,  # Increase verbosity.
    'delete-cache': None,  # Delete local cache files.
    'sync': None,  # Pulls updates from the server, but does not push changes to the server.
    'commit': None,  # Push changes to the server.
    'show-queue': None,  # Show the command queue, that will be pushed to the server on `commit`.
}

# These are actions that requires the full `api` object to work,
# e.g. because they need to convert a project-name to project-id.
API_ACTIONS = {
    "add-task": add_task,
}



# Copyright 2019, Rasmus Sorensen <rasmusscholer@gmail.com>
"""

Module with all value transformation functions.

E.g., if you specify `value_transform=int` to `filter_tasks()`, then the comparison
value will be converted using `value = int(value)`.

This module just imports all value transformation functions into a single module/namespace.

"""

from builtins import *  # provides int, float, eval (!), etc.
from datetime import time, datetime
from dateutil.parser import parse as parse_dateutil
from dateparser import parse as parse_dateparser

parse_date = parse_dateutil







# Copyright 2018 Rasmus Scholer Sorensen

"""

This is a module to create generic query/filter operators.

It is pretty basic, a query consists of expressions (currenly joined by AND).

Each expressions is of the form "value1 OPERATOR value2", e.g.
    string1 equals string2
    'abcdefg' startswith 'abc'  # True
    'abcdefg' contains 'cde' # True
    etc.


"""

from re import match as _match  # Wrong signature: re.match(pattern, string), I want "string re pattern"
from fnmatch import fnmatch, fnmatchcase

from operator import eq, ne, lt, le, gt, ge, mod, countOf, contains, indexOf, and_, or_, is_, is_not
# unary operators: not_, truth, is_not

# Aliases:
equals = equal = eq
neq = nequal = ne
less = lessthan = lt
greater = greaterthan = gt
modulus = mod
glob = fnmatchcase


def to_lower(obj):
    """ Produces a lower-cased representation of obj.
    If obj is a list or set, returns a list or set where `to_lower()`
    has been applied to all elements.

    If obj is a dict, returns a list or set where `to_lower()`
    has been applied to all keys and values.

    Else, return `str(obj).lower()`
    """
    if isinstance(obj, (set, list)):
        return type(obj)([to_lower(val) for val in obj])
    elif isinstance(obj, dict):
        return type(obj)([(to_lower(key), to_lower(val)) for key, val in obj.items()])
    else:
        return str(obj).lower()


def startswith(a, b):
    a = str(a)
    b = str(b)
    return a.startswith(b)


def istartswith(a, b):
    """ Case insensitive startswith. """
    a, b = to_lower(a), to_lower(b)
    return a.startswith(b)


def endswith(a, b):
    a = str(a)
    b = str(b)
    return a.endswith(b)


def iendswith(a, b):
    """ Case insensitive startswith. """
    a, b = to_lower(a), to_lower(b)
    return a.endswith(b)


def icontains(a, b):
    """ Case insensitive contains.
    Note that the order of operands is reversed between `contains` and `in_`:

        contains(a, b)  <=>   b in a

    """
    a, b = to_lower(a), to_lower(b)
    return contains(a, b)


def in_(a, b):
    """ Returns True if `a` contains `b`, i.e. `contains` but with args in reversed order.

    Note that the order of operands is reversed between `contains` and `in_`:

        contains(a, b)  <=>   b in a    <=> in_(b, a)

    OBS: Please make sure this is really what you mean to use!
    The general format of the `-filter` action is:
        `-filter <key> <op> <value>`,
    e.g.
        `-filter content in "task1_content task2_content".
    """
    return a in b


def iin(a, b):
    """ Returns True if `a.lower()` contains `b.lower()`, i.e. `contains` but with args in reversed order."""
    a, b = to_lower(a), to_lower(b)
    return a in b


def ieq(a, b):
    """ Case insensitive equality. """
    a, b = to_lower(a), to_lower(b)
    return a == b


def ine(a, b):
    """ Case insensitive in-equality. """
    a, b = to_lower(a), to_lower(b)
    return a != b


def ilt(a, b):
    """ Case insensitive less-than. """
    a, b = to_lower(a), to_lower(b)
    return a < b


def igt(a, b):
    """ Case insensitive greater-than. """
    a, b = to_lower(a), to_lower(b)
    return a > b


def ige(a, b):
    """ Case insensitive greater-than-or-equal-to. """
    a, b = to_lower(a), to_lower(b)
    return a > b


# def is_(a, b):
#     return a is b


def re(a, b):
    """ Returns True if `a` matches the regular expression given by `b`.
    Unfortunately, the call signature of re.match(pat, name) is different from the other binary operators,
    which are usually `a lt b  ==> lt(a, b) ==>  a < b`,
    Note that `fnmatchcase(name, pat)` follows the "correct" order.
    """
    return _match(b, a)


def ire(a, b):
    """ Returns True if `a.lower()` matches the regular expression given by `b.lower()`."""
    a, b = to_lower(a), to_lower(b)
    return _match(b, a)


def ifnmatch(a, b):
    """ Case insensitive glob-style matching. """
    a, b = to_lower(a), to_lower(b)
    return fnmatchcase(a, b)


iglob = ifnmatch



# Copyright 2018-2019 Rasmus Scholer Sorensen <rasmusscholer@gmail.com>

"""

Module for task-specific functionality, e.g. adding extra fields to tasks, etc.
Functions must operate on todoist.Item objects, or the equivalent dicts (or a list of tasks).


Note:
ISO 8601 also defines a format for durations: PnYnMnDTnHnMnS
e.g. P3Y6M4DT12H30M5S, or P2H35M - similar to just writing 2h30m, except the ISO8601 uses capitals for time.

The Todoist datetime string format is "almost" the "ctime" format:

except that Todoist has year before time of day, while ctime has year at the end.

"""

import sys
import datetime
import re
import dateutil.parser
import pytz
from dateutil import tz
from todoist.models import Item, Project, Label
from copy import deepcopy
from pprint import pprint

# Note: To get localized date formats, use the "Babel" package, c.f. https://stackoverflow.com/a/32785195/3241277
ISO_DATE_FMT = "%Y-%m-%dT%H:%M:%S"
DATE_TIME_FMT = "%Y-%m-%d %H:%M"  # Prettier format than ISO8601
DATE_NO_TIME_FMT = "%Y-%m-%d"
LABEL_REGEX = r"@\w+"
EXTRA_PROPS_REGEX = r"(?P<prop_group>\{(?P<props>(\w+:\s?[^,]+,?\s?)*)\})"
PROP_KV_REGEX = r"((?P<key>\w+):\s?(?P<val>[^,]+),?\s?)"
TASK_REGEX = r"^(?P<title>.*?)" + EXTRA_PROPS_REGEX + r"*\s*$"
# extra_props_regex = r"(?P<prop_group>\{(R|r)eward:\s?(?P<reward>.+)\})"
# extra_props_regex = r"(?P<prop_group>\{(?P<props>\w+:\s?[^,]+)*\})"
# prop_kv_regex = r"(?P<key>\w+):\s?(?P<val>[^,]+)"
# TASK_REGEX = r"^(?P<title>.*?)\s*(?P<reward_group>\{(R|r)eward:\s?(?P<reward>.+)\})?\s*$"
LOCAL_TIMEZONE = tz.tzlocal()
NO_DUEDATE_DATETIME = datetime.datetime(2099, 12, 31, 23, 59, 59).astimezone(LOCAL_TIMEZONE)
END_OF_DAY_TIME = datetime.time(23, 59, 59)
NO_DUE_DATE_PRETTY_STR = "(No due-date)"


def get_proper_priority_int(priority) -> int:
    """ Return a correct priority integer value for use with the API.

    The input priority can be either a string ("p1", "p2", "p3", or "p4") where "p1" is
    highest priority and "p4" is lowest priority. This is converted to one of the values
    (4, 3, 2, or 1), where priority=4 is highest priority.

    Alternatively, the input priority can be an integer, in which case we simply check that the
    integer is either 1, 2, 3, or 4, where priority=4 is highest priority.

    Examples:
        priority = "p3"
        priority = get_proper_priority_int(priority)
        assert priority == 2

        priority = 1
        priority = get_proper_priority_int(priority)
        assert priority == 1

        priority = 0
        priority = get_proper_priority_int(priority)
        ValueError: Argument `priority` must be between 1 and 4 (if int).

        priority = None
        priority = get_proper_priority_int(priority)
        TypeError: Argument `priority` must be str or int.
    """
    if isinstance(priority, str):
        # Priority in the form of "p1" to "p4".
        priority_str = priority
        priority_str_map = dict(p1=4, p2=3, p3=2, p4=1)
        try:
            priority = priority_str_map[priority]
        except KeyError:
            raise ValueError("Argument `priority` must be one of 'p1', 'p2', 'p3', or 'p4' (if str).")
    else:
        # Integer between 1 and 4:
        if not isinstance(priority, int):
            raise TypeError("Argument `priority` must be str or int.")
        if priority < 1 or 4 < priority:
            raise ValueError("Argument `priority` must be between 1 and 4 (if int).")
    return priority


def get_task_data(task, data_attr="_custom_data"):
    if isinstance(task, Item):
        # Try the "_custom_data" first (it should include original data as well), fall back to Item.data:
        return getattr(task, data_attr, task.data)
    else:
        return task


def get_task_value(task, taskkey, default=None, coerce_type=None, data_attr="_custom_data"):
    """ Get a task data value from the task.
    This function combines a couple of features:

    1. Determine if task is a dict or Item, and extract the data dict using the given data attribute.
    2. Get the data value given by taskkey, defaulting to the given default value
        if `key` is not present in the data dict.
    3. coerce the value to a given type, e.g. coerce "1" to integer 1, or a date-string to datetime object.

    Args:
        task: Either a todoist.models.Item, or a task dict.
        taskkey: The key whose value you would like to get.
        default: Default value, if no value could be found.
        coerce_type: Coerce the value to this type before returning.
        data_attr: The attribute of task where data is stored.
            I am storing custom/derived data in task._custom_data, separate from the original
            data, to prevent the derived data from being cached to disk.

    Returns:
        Value, or <default>

    """
    task_data = get_task_data(task, data_attr=data_attr)
    try:
        task_value = task_data[taskkey]
    except KeyError:
        if taskkey.startswith("due_"):
            try:
                task_value = task_data['due'][taskkey.strip("due_")]
            except (KeyError, TypeError):  # TypeError if task_data['due'] is None
                task_value = default
        elif '.' in taskkey:
            # support dot-notation, task_data["due.date"] -> task_data["due"]["date"]
            basekey, childkey = taskkey.split('.', 1)
            try:
                task_value = get_task_value(task_data[basekey], childkey)
            except (KeyError, TypeError):
                task_value = default
        else:
            task_value = default
    if coerce_type is not None and task_value is not None and type(task_value) != coerce_type:
        print("NOTICE: `type(task_value) != coerce_type` - Coercing `task_value` to %s:" % type(coerce_type))
        task_value = coerce_type(task_value)
    return task_value


def is_recurring(task, data_attr='_custom_data'):
    """ Return whether task is recurring or not.
    This is super easy for new v8 Sync API tasks, but not so much for the older tasks.
    """
    task_data = get_task_data(task)
    try:
        return task_data['due']['is_recurring']
    except (KeyError, TypeError):  # TypeError if task_data['due'] is None
        due_string = task_data.get('date_string') or task_data.get('due_string') or ''
        return 'every' in due_string


def get_recurring_tasks(tasks: list, negate: bool = False) -> list:
    return [task for task in tasks if is_recurring(task) != negate]


def extract_labels(content):
    labels = re.findall(LABEL_REGEX, content)
    cleaned = re.sub(LABEL_REGEX, "", content).rstrip()
    return labels, cleaned


def extract_props(content):
    props_match = re.search(EXTRA_PROPS_REGEX, content)
    if not props_match:
        return None, content
    props_str = props_match.groupdict()['props']
    props_pairs = re.findall(PROP_KV_REGEX, props_str)
    props_dict = dict(tuple(mgroup[1:3]) for mgroup in props_pairs)

    cleaned = re.sub(EXTRA_PROPS_REGEX, "", content).rstrip()
    return props_dict, cleaned


CUSTOM_FIELDS = {
    # Update: There was too many date fields, so now all timestamps and datetime objects are just in local time.
    'due_date',  # Due date (str), as listed by the API.
    'due_date_string',  # Human-readable due date, as listed by the API.
    'due_string_safe',  # Save due_string, is None if no due date.
    'due_date_dt',  # datetime object (local time)
    'due_date_iso',  # ISO str timestamp (local time)
    'due_date_safe_dt',  # Safe due date datetime - is always present (dummy value for no due date)
    'due_date_safe_iso',  # Safe due date ISO str - is always present (dummy value for no due date)
    # 'due_date_utc_iso',
    # 'due_date_local_dt',
    # 'due_date_local_iso',
    'date_added_dt',   # datetime object (local time)
    'date_added_iso',   # ISO str timestamp (local time)
    'date_added_safe_dt',
    'date_added_safe_iso',
    # 'date_added_utc_iso',
    # 'date_added_local_dt',
    # 'date_added_local_iso',
    'completed_date_dt',
    'completed_date_iso',
    'completed_date_safe_dt',
    'completed_date_safe_iso',
    # 'completed_date_utc_iso',
    # 'completed_date_local_dt',
    # 'completed_date_local_iso',
    'checked_str',
}


def add_task_date_fields(
        input_dict, output_dict=None,
        date_keys=("date_added", "date_completed", "completed_date"),
        allday_time=END_OF_DAY_TIME,
        safe_date=NO_DUEDATE_DATETIME,
        datestrfmt="%Y-%m-%d %H:%M",
        allday_datestrfmt="%Y-%m-%d",
):
    """ Parse dates in `input_dict` and add custom date fields to `output_data`.

    Args:
        input_dict:
        output_dict:
        allday_time: If specified, "all-day" tasks (with a due date but no due time)
            will have their "due_time_dt" datetime object set to this value.
        safe_date: If a date is not specified (due date, date added, date completed),
            use this as the "safe value", e.g. typically a date really far into the future.

    Returns:
        output_dict

    Update, 2019: The new v8 Sync API builds on the combined `due` property that was introduced
    with v7.1.

    Examples:
        >>> task._custom_data = {}  # Create custom data dict.
        >>> output = add_task_date_fields(input_dict=task.data, output_dict=task._custom_data)

    """
    if output_dict is None:
        output_dict = {}

    # Parse due date:
    if input_dict.get('due'):
        # v8 (and v7.1) Sync API has separate 'due' child dict - expand these to the old v7.0 format:
        output_dict['due_date'] = input_dict['due']['date']
        output_dict['due_string'] = input_dict['due']['string']
        output_dict['due_string_safe'] = input_dict['due']['string']
        output_dict['date_string'] = input_dict['due']['string']  # v7.0 legacy attribute name
        output_dict['due_is_recurring'] = input_dict['due']['is_recurring']
        output_dict['is_recurring'] = input_dict['due']['is_recurring']
        # An all-day task in v8 API has a due date with no time specification:
        output_dict['is_allday'] = len(input_dict['due']['date']) <= len("YYYY-MM-DD")
        # Note: v8 due date is in RFC-3339/ISO8601 format!

        # Parse strings to datetime objects (denoted with a "_dt" postfix):
        # v8 API has two distinct ways of specifying due dates:
        # 1. "Floating", where the time is always the literal clock time, in-dependent of timezone.
        # 2. "Fixed timezone", where the time is in universal UTC time, and has a "timezone" property.
        #    The timezone attribute is used when re-parsing the due-string,
        # e.g. string="every day at 2pm", timezone="Europe/Copenhagen".
        # OBS: The datetime objects are always represented in the local timezone.
        output_dict['due_date_dt'] = dateutil.parser.parse(input_dict['due']['date'])
        if output_dict['is_allday'] and allday_time:
            # Force v7.0 behaviour for datetime objects:
            output_dict['due_date_dt'] = output_dict['due_date_dt'].replace(
                hour=allday_time.hour, minute=allday_time.minute, second=allday_time.second
            )
            output_dict['due_date_dt'] = output_dict['due_date_dt'].astimezone(tz.tzlocal())
        if not output_dict['due_date_dt'].tzinfo:
            if input_dict['due']['timezone']:  # Only check timezone if not all-day task:
                # OBS: Timezone calculation not needed for v7.1 legacy tasks, where [due][date] was UTC timestamp:
                # Convert from the given timezone to local timezone.
                # First create a "due-date local" timezone object, then use it to make a timezone-aware
                # datetime object using `localize()`, then convert that timezone-aware datetime to
                # computer-local time using `astimezone(tz.tzlocal())`:
                # print("Task '{content}': due: {due} => dt: {due_date_dt}".format(due_date_dt=output_dict['due_date_dt'], **input_dict))
                # Make it aware of the timezone:
                output_dict['due_date_dt'] = pytz.timezone(input_dict['due']['timezone']).localize(output_dict['due_date_dt'])
            # Convert from UTC (or whatever timezone it has) to local datetime:
            output_dict['due_date_dt'] = output_dict['due_date_dt'].astimezone(tz.tzlocal())
            # output_dict['due_date_dt'] = pytz.utc.localize()
        output_dict['due_date_pretty_safe'] = output_dict['due_date_dt'].strftime(
            DATE_NO_TIME_FMT if output_dict['is_allday'] else DATE_TIME_FMT)
    elif input_dict.get('due_date_utc'):
        # I have some old tasks where "due_date_utc" attribute is present but set to None. (weird)
        # print("Task '{content}': parsing due_date_utc: {due_date_utc}   (string: {date_string})".format(**input_dict))
        # Old v7 Sync API
        output_dict['due_date'] = input_dict['due_date_utc']
        output_dict['due_string'] = input_dict['date_string']
        output_dict['due_string_safe'] = input_dict['date_string']
        # xx:xx:59 = due date with no time (v7.0)
        output_dict['is_allday'] = (input_dict['due_date_utc'] or "59")[-2:] == "59"
        output_dict['due_date_dt'] = dateutil.parser.parse(input_dict['due_date_utc']).astimezone(LOCAL_TIMEZONE)
        output_dict['due_date_pretty_safe'] = output_dict['due_date_dt'].strftime(
            DATE_NO_TIME_FMT if output_dict['is_allday'] else DATE_TIME_FMT)
    else:
        # No due date specified:
        output_dict['is_allday'] = True
        output_dict['due_string_safe'] = input_dict.get('date_string', None)
        output_dict['due_date_pretty_safe'] = NO_DUE_DATE_PRETTY_STR

    # All datetime objects should have a timezone (either that, or none should have timezone)
    # Mixing timezone-aware with timezone-naive makes comparison impossible (TypeError)
    # At this point, all `due_date_dt` objects should have a timezone. If not, then something is wrong.
    if output_dict.get('due_date_dt') and not output_dict['due_date_dt'].tzinfo:
        print("Adding tzinfo to task (it should already have it at this point!):")
        pprint(input_dict)
        output_dict['due_date_dt'] = output_dict['due_date_dt'].astimezone(tz.tzlocal())

    # Add some "guaranteed", safe fields which are always present:
    output_dict['due_date_safe_dt'] = output_dict.get('due_date_dt', safe_date)
    output_dict['due_date_safe_iso'] = "{:%Y-%m-%dT%H:%M:%S}".format(output_dict.get('due_date_dt', safe_date))
    output_dict['due_date_safe'] = output_dict.get('due_date', "(No due-date)")

    # Parse additional date attributes to datetime objects (with local timezone):
    for key in date_keys:
        datestr = input_dict.get(key)
        # Old v7 date format was e.g. "Fri 23 Mar 2018 15:01:05 +0000" - ridiculous.
        # New v8 format is strictly RFC-3339/ISO1806 - nice.
        if datestr:
            # Create datetime object - datestr should be guaranteed to be UTC:
            dt_local = dateutil.parser.parse(datestr).astimezone(LOCAL_TIMEZONE)
            output_dict['%s_dt' % key] = dt_local
            output_dict['%s_iso' % key] = "{:%Y-%m-%dT%H:%M:%S}".format(dt_local)
        else:
            dt_local = safe_date.astimezone(LOCAL_TIMEZONE)

        # It is nice to have some "guaranteed", safe fields which are always present:
        output_dict['%s_safe_dt' % key] = dt_local
        output_dict['%s_safe_iso' % key] = "{:%Y-%m-%dT%H:%M:%S}".format(dt_local)

    # Also make "checked_str" field to quickly indicate the checked/completed status of a task:
    output_dict["checked_str"] = "[x]" if input_dict.get("checked", 0) else "[ ]"
    # Also make priority string, "p1", "p2". This is kind of weird, because p1 (high priority) is 4 not 1.
    # (The "priority strings" are used e.g. in the web app, where you type "p1" to make a high-priority task.)
    output_dict["priority_str"] = "p%s" % (5 - input_dict.get("priority", 1))
    return output_dict


def get_input_output_dicts(task, output_attr="_custom_data", deepcopy_data=True):
    if isinstance(task, Item):
        input_data = task.data
        if output_attr:
            try:
                output_data = getattr(task, output_attr)
            except AttributeError:
                output_data = deepcopy(task.data) if deepcopy_data else {}
                setattr(task, output_attr, output_data)
        else:
            output_data = task.data
    else:
        assert isinstance(task, dict)
        input_data = output_data = task
    return input_data, output_data


def inject_tasks_custom_data(
        tasks, output_attr="_custom_data", deepcopy_data=True,
        add_dates=True,
        parse_content=False, task_regex=TASK_REGEX,
        add_project_info=True, projects=None
):
    """ Parse task data (e.g. dates) and inject custom data fields.

    This an "all-in-one" function that performs all parsing/extraction steps.
    (Not sure this is better that just doing them sequentially.)

    By default, the custom data is added to a separate `task._custom_data` property,
    which will be a deep copy of `task.data`.

    Args:
        tasks:
        output_attr: The task attribute to inject data to.
        deepcopy_data:
        add_dates:
        parse_content:
        task_regex:
        add_project_info:
        projects:

    Returns:

    """
    for task in tasks:
        # Get the input_data and output_data objects to read from and write to:
        input_data, output_data = get_input_output_dicts(
            task=task, output_attr=output_attr, deepcopy_data=deepcopy_data)

        if add_dates:
            # Custom date fields (and "checked_str" and "priority_str"):
            add_task_date_fields(input_dict=input_data, output_dict=output_data)
            # output_data.update(custom_fields)

        if parse_content:
            parse_task_content(task, output_data, task_regex)

    # Adding project info requires having the projects data:
    if add_project_info:
        assert projects is not None
        inject_tasks_project_fields(tasks=tasks, projects=projects)


def inject_tasks_date_fields(
        tasks,
        date_keys=("date_added", "date_completed", "completed_date"),
        strict=False,
        output_attr='_custom_data', deepcopy_data=True,
        verbose=0
):
    """ Parse date strings and create python datetime objects.


    """
    if verbose:
        print(f"Parsing and adding additional date information to tasks {output_attr if output_attr else ''}...",
              file=sys.stderr)
    for task in tasks:
        # Get the input_data and output_data objects to read from and write to:
        input_data, output_data = get_input_output_dicts(
            task=task, output_attr=output_attr, deepcopy_data=deepcopy_data)

        # Add custom date fields (and "checked_str" and "priority_str"):
        added_fields = add_task_date_fields(
            input_dict=input_data, output_dict=output_data, date_keys=date_keys)
        # print(f"Task {input_data['content']} aded date fields:", added_fields)
        # output_data.update(added_fields)


def inject_tasks_project_fields(tasks, projects, strict=False, na='N/A', output_attr='_custom_data', verbose=0):
    """ Inject project information (name, etc) for each task, using the task's project_id.
    This makes it considerably more convenient to print tasks, when each task has the project name, etc.
    Information is injected as `getattr(task, output_attr, task.data)["project_%s" % k"] = v`
    for all key-value pairs in the task's project, e.g. project_name, project_color, project_indent,
    project_is_archived, etc.

    Args:
        tasks: A list of tasks.
        projects: A list of projects, or a dict of projects keyed by `project_id`.
        strict: If True, raise an error if a task's project_id is not found in the projects dict.
        na: If, in non-strict mode, a task's project_id is not found, use this value instead.
            It can be either a single value, which is used for the most common project fields,
            or it can be a dict where we just do `task.update(na)`.
        output_attr: Instead of updating task directly, update the dict found
            in `getattr(task, output_attr)`.

    Returns:
        None; the task dicts are updated in-place.
    """
    if verbose:
        print(f"Adding additional project information to tasks {output_attr if output_attr else ''}...",
              file=sys.stderr)
    if not isinstance(projects, dict):
        # If projects is e.g. a list of projects, create a dict mapping project_id to project:
        projects_by_pid = {project['id']: project for project in projects}
        projects = projects_by_pid
    # I think it should be OK to add non-standard fields to task Items
    for task in tasks:
        # Update either `task.data` or `task._custom_data`:
        input_data, output_data = get_input_output_dicts(task, output_attr=output_attr, deepcopy_data=True)
        pid = input_data['project_id']
        # Todoist API sometimes returns string ids and sometimes integer ids.
        try:
            project = projects[pid if pid in projects else str(pid)]
        except KeyError as exc:
            if strict:
                raise exc
            else:
                if isinstance(na, dict):
                    output_data.update(na)
                else:
                    output_data['project_name'] = na
        else:
            # Add all the project info to task, using "project_" prefix.
            # If project is a Model instance, then the data dict is in the 'data' attribute
            # (otherwise just project is already a dict and can be used directly).
            # (the todoist.model.Model class does not support the dict interface).
            for k, v in getattr(project, 'data', project).items():
                output_data["project_%s" % k] = v


def inject_tasks_labels_fields(
        tasks, labels, output_attr='_custom_data', label_fmt="@{name}", labels_sep=" ", verbose=0):
    """ Inject labels information for each task, using the task's 'labels' attribute.
    This makes it considerably more convenient to print tasks, when each task has the label name,
    rather than just the label id.

    Args:
        tasks: A list of tasks.
        labels: A list of labels, or a dict of labels keyed by `label_id`.
        output_attr: Instead of updating task directly, update the dict found
            in `getattr(task, output_attr)`.
        label_fmt: How to format each label when creating `labels_str`.
        labels_sep: How to join the labels when creating `labels_str`.

    Returns:
        None; the task dicts are updated in-place.
    """
    if verbose:
        print(f"Adding additional labels information to tasks {output_attr if output_attr else ''}...",
              file=sys.stderr)
    if not isinstance(labels, dict):
        # If projects is e.g. a list of projects, create a dict mapping project_id to project:
        labels_by_id = {int(label['id']): label for label in labels}
    else:
        labels_by_id = labels
    # I think it should be OK to add non-standard fields to task Items
    del labels
    for task in tasks:
        # Update either `task.data` or `task._custom_data`:
        input_data, output_data = get_input_output_dicts(task, output_attr=output_attr, deepcopy_data=True)
        task_labels = [labels_by_id[lid] for lid in input_data['labels']]
        label_dicts = [label.data if isinstance(label, Label) else label for label in task_labels]
        output_data['label_names'] = [label['name'] for label in label_dicts]
        # if lowercase_label_names:
        #     output_data['label_names'] = [label_name.lower() for label_name in output_data['label_names']]
        output_data['labels_str'] = labels_sep.join(label_fmt.format(**label) for label in label_dicts)


def parse_task_content(task, output_dict, task_regex):
    try:
        output_dict.update(re.match(task_regex, task['content']).groupdict())
        labels, cleaned = extract_labels(task['content'])
        props, cleaned = extract_props(task['content'])
        output_dict['cleaned'] = cleaned
        output_dict['ext_labels'] = labels
        output_dict['ext_props'] = props or {}
    except AttributeError as e:
        print("WARNING: Error while matching regex `{}` to task['content'] `{}`: %s".format(
            task_regex, task['content'], repr(e)))
    return output_dict


def parse_tasks_content(tasks, task_regex=None, output_attr='_custom_data', verbose=0):
    """ Parse tasks using the task-parsing regular expressions. Tasks are parsed and updated in-place.
    This is only for use with my custom metadata scheme where I use `{reward: 1h}` in the task content
    to define key-value metadata pairs.

    ^(?P<title>.*?)\s*(?P<reward_group>\{reward:\s?(?P<reward>.+)\})
    (?P<title>.*)\s*(?P<reward_group>\{reward:\s?(?P<reward>.+)\})
    ^(?P<title>.*?)\s*(?P<reward_group>\{reward:\s?(?P<reward>.+)\})?\s*$
    ^(?P<title>.*?)\s*(?P<reward_group>\{(R|r)eward:\s?(?P<reward>.+)\})?\s*(?P<labels>(@\w+\s?)*\s?)*$
    ^(?P<title>.*?)\s*(?P<prop_group>\{(R|r)eward:\s?(?P<reward>.+)\})?\s*(?P<labels>(@\w+\s?)*)\s*$

    Check DFCI email {reward: 0.25h W} @Habit @Reward
    Check DFCI email @Habit @Reward {reward: 0.25h W}
    @Habit @Reward Check DFCI email {reward: 0.25h W}

    """
    if verbose:
        print(f"Adding additional content-parsed information to tasks {output_attr if output_attr else ''}...",
              file=sys.stderr)
    if task_regex is None:
        task_regex = TASK_REGEX
    if isinstance(task_regex, str):
        task_regex = re.compile(task_regex)
    for task in tasks:
        # Update either `task.data` or `task._custom_data`:
        if output_attr:
            try:
                output_data = getattr(task, output_attr)
            except AttributeError:
                output_data = {}
                setattr(task, output_attr, output_data)
        else:
            output_data = task
        parse_task_content(task, output_data, task_regex)
    return tasks




# Copyright 2018 Rasmus Scholer Sorensen

"""

The main Todoist v7 "Sync" API has become rather complex.
That is probably required for "full-featured" applications,
but for small command-line scripts, the Sync API is a bit overkill.

Doist is therefore (as of primo 2018) introducing a new, simple, REST API,
which can be used for those simple use-cases.


This module features classes and functions to utilize the Todoist REST API.

Update: The REST API is out.

* It is based on the v8 Sync API specs (e.g. separate `due` property), but is called "REST v1".
* Although creating a new Task Item actually specifies passing `due_string` and `due_lang` values.
    * https://developer.todoist.com/rest/v1/?python#create-a-new-task
* https://developer.todoist.com/rest/v1/?python

"""

import requests

from actionista.todoist.config import get_token


class TodoistRestApi:
    """ A simple class for interacting with the Todoist REST API, currently at https://beta.todoist.com/API/v8/. """

    API_BASE_URL = "https://api.todoist.com/rest/v1/"


    def __init__(self, token=None):
        self._token = None
        if token is None:
            token = get_token()
        self.update_token(token)
        self.session = requests.Session()

    def update_token(self, token):
        self._token = token
        self.session.headers.update({
            "Authorization": "Bearer %s" % token})

    def get(self, endpoint, **kwargs):
        endpoint_url = self.API_BASE_URL + endpoint
        res = self.session.get(url=endpoint_url, **kwargs)
        res.raise_for_status()
        return res.json()

    def post(self, endpoint, **kwargs):
        endpoint_url = self.API_BASE_URL + endpoint
        res = self.session.post(url=endpoint_url, **kwargs)
        res.raise_for_status()
        return res.json()

    def delete(self, endpoint, **kwargs):
        endpoint_url = self.API_BASE_URL + endpoint
        res = self.session.delete(url=endpoint_url, **kwargs)
        res.raise_for_status()
        return res.json()

    def get_tasks(self, project_id=None, label_id=None, filter=None):
        kwargs = dict(project_id=project_id, label_id=label_id, filter=filter)
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return self.get(endpoint='tasks', **kwargs)

    def close_task(self, task_id: int):
        assert isinstance(task_id, int)
        endpoint = f"tasks/{task_id}/close"
        res = self.post(endpoint=endpoint)
        print(f"Closed task {task_id}; response: {res}  (should be 204).")
        return res

    def delete_task(self, task_id: int):
        assert isinstance(task_id, int)
        endpoint = f"tasks/{task_id}"
        res = self.delete(endpoint=endpoint)
        print(f"Deleted task {task_id}; response: {res}  (should be 204).")
        return res

    def get_task_comments(self, task_id: int):
        assert isinstance(task_id, int)
        endpoint = f"comments/"
        params = {"task_id": task_id}
        return self.get(endpoint=endpoint, params=params)





"""

Module dealing specifically with dates and times.

"""
import datetime

import parsedatetime
from dateutil import tz
import dateparser
# import datetime
# import pendulum

DATE_DAY_FMT = '%Y-%m-%d'
ISO_8601_FMT = '%Y-%m-%dT%H:%M:%S'
TODOIST_DATE_FMT = "%a %d %b %Y %H:%M +0000"  # e.g. 'Mon 07 Aug 2006 12:34:56 +0000'. Use xx:xx:59 as "all day" time.
# Todoist date format is similar to RFC2822, except that it doesn't have a comma after the weekday.


def utc_time_to_local(utcdatetime, timezone=None, fmt="datetime"):
    """ Todoist has some serious time issues. All due times are stored and transmitted in UTC time.
    But, "special times", e.g. 23:59:59 to mark "end of day" or "all day", is in *local time*.
    Furthermore, there is e.g. summer vs winter time. So, before daylight saving time,
    "end of day" tasks may be 22:59:59 UTC, but after DST "end of day" may be marked as 21:59:59 UTC
    because both times will be 23:59:59 local time.
    You thus need the date in order to convert times between timezones.
    Libraries to deal with datetime conversions:
        pendulum
        zulu
        dateutil
        datetime

    Examples:
        >>> utcdatetime = "Sat 24 Mar 2018 22:59:59 +0000"
        >>> utc_time_to_local(utcdatetime)
        >>> utcdatetime = "Mon 26 Mar 2018 21:59:59 +0000"
        >>> utc_time_to_local(utcdatetime)
    """
    if timezone is None:
        timezone = tz.tzlocal()
    if isinstance(utcdatetime, str):
        utcdatetime = dateparser.parse(utcdatetime)
    if utcdatetime.tzinfo is None:
        utcdatetime.replace(tzinfo=tz.gettz('UTC'))

    # Convert UTC time to local:
    localdt = utcdatetime.astimezone(timezone)

    if fmt is None or fmt == "datetime":
        return localdt
    elif fmt == "date":
        return localdt.date()
    elif fmt == "time":
        return localdt.time()
    else:
        return localdt.strftime(fmt)


def local_time_to_utc(localtime, timezone=None, fmt="datetime"):
    if timezone is None:
        timezone = tz.tzlocal()
    if isinstance(localtime, str):
        localtime = dateparser.parse(localtime)
    if localtime.tzinfo is None:
        localtime.replace(tzinfo=timezone)

    # Convert local time to UTC:
    utcdatetime = localtime.astimezone(tz.gettz('UTC'))

    if fmt is None or fmt == "datetime":
        return utcdatetime
    elif fmt == "date":
        return utcdatetime.date()
    elif fmt == "time":
        return utcdatetime.time()
    else:
        return utcdatetime.strftime(fmt)


def end_of_day(dt):
    return dt.replace(hour=23, minute=59, second=59)


def start_of_day(dt):
    return dt.replace(hour=0, minute=0, second=0)


def human_date_to_iso(human_date, fmt=ISO_8601_FMT):
    """ Convert natural language / contextual dates, e.g. 'today', to iso date string.
    Note: This is a little bit finicky. For instance, "in 2 weeks" works, but none of the following variations do:
        "in two weeks", "2 weeks from now", "two weeks", "in 2 weeks from now", etc.

    The returned datetime object does not by default have any timezone information,
    so the iso datestring is basically local time.
    """
    dt = dateparser.parse(human_date)
    if dt is None:
        print(f"\nERROR: FAILED to parse human input date {human_date!r}.\n")
        raise ValueError(f"FAILED to parse human input date {human_date!r}.")
    return dt.strftime(fmt)


def get_rfc3339_datestr(datestr: str):
    try:
        datetime.datetime.fromisoformat(datestr)  # as of Python 3.7
    except ValueError:
        # If datetime.datetime.fromisoformat() is unable to parse the date_str,
        # then it is probably not an iso string and we should re-parse it.
        # Note that dateutil.parser.parse("monday") works just fine.
        cal = parsedatetime.Calendar()
        time_struct, context = cal.parse(datestr, version=2)
        dt = datetime.datetime(*time_struct[:6])
        if context.hasTime:
            date_rfc3339 = dt.isoformat()
        else:
            date_rfc3339 = f"{dt:%Y-%m-%d}"
    else:
        date_rfc3339 = datestr
    return date_rfc3339



"""
Actionista Action CLI for Todoist (actionista-todoist).

A `find`-inspired CLI for Todoist.

Add, select, print, reschedule, modify, and complete tasks in a batch-wise fashion
from the command line.

"""

# Do not zero-pad month or day:
__version__ = '2019.9.10'  # remember to also update setup.py



# Copyright 2019, Rasmus Sorensen <rasmusscholer@gmail.com>
"""

Dedicated add-item CLI.

This is a newer version of the "Ad-hoc CLIs" found in `adhoc_cli.py`.

The CLIs found here are based on `click`, while `adhoc_cli` uses `argparse`.

"""

import sys
import click
from pprint import pprint, pformat

from .utils import get_todoist_api
from .tasks_utils import get_proper_priority_int
from .config import get_config
from .config import DEFAULT_PROJECT_PRINT_FMT, DEFAULT_PROJECT_SORT_KEYS, DEFAULT_PROJECT_SORT_ORDER


@click.group("todoist-cli")
def todoist_cli():
    """ Actionista for Todoist CLI.

    Note: This is NOT the "Action Chain CLI", but a more traditional CLI, grouping
    the following CLI commands:

    See also: `todoist-adhoc` CLI command.
    """
    pass

# todoist_cli.add_command(add_task)  # Add manually (without decorator).

# Q: Can this be invoked directly when decorated with Group.command(), instead of click.command()?
# A: Yes, `add_task_cli` can still be invoked directly.
@todoist_cli.command("add-task")  # NOT click.command().
@click.argument("content")
@click.option("--due", metavar="DUE-DATE")
@click.option("--project", metavar="PROJECT-NAME")
@click.option("--label", "labels", multiple=True, metavar="LABEL")
@click.option("--priority", type=click.Choice(["p1", "p2", "p3", "p4"]))
@click.option("--note", metavar="NOTE")
@click.option("--auto-reminder/--no-auto-reminder", default=True)
@click.option("--auto-parse-labels/--no-auto-parse-labels", default=True)
@click.option("--sync/--no-sync", default=True)
@click.option("--commit/--no-commit", default=True)
@click.option("--show-queue/--no-show-queue", default=False)
@click.option("--verbose", "-v", count=True)
def add_task_cli(
        content, due=None, project=None, labels=None, priority=None, note=None,
        auto_reminder=None, auto_parse_labels=None,
        sync=True, commit=True, show_queue=False, verbose=0,
):
    """ Actionista for Todoist: Add-task CLI """
    return add_task(
        content, due=due, project=project, labels=labels, priority=priority, note=note,
        auto_reminder=auto_reminder, auto_parse_labels=auto_parse_labels,
        sync=sync, commit=commit, show_queue=show_queue,
        verbose=verbose,
    )


def add_task(
        content, due=None, project=None, labels=None, priority=None, note=None,
        auto_reminder=None, auto_parse_labels=None,
        sync=True, commit=True, show_queue=False, verbose=0,
):
    """ Add a single task to Todoist.

    Args:
        content: The task content (task name).
        due: Due date (str)
        project: Project name (str)
        labels: List of task labels (names)
        priority: Task priority, either number [1-4] or str ["p4", "p3", 'p2", "p1"].
        note: Add a single note to the task.
        auto_reminder: Automatically add a default reminder to the task, if task is due at a specific time of day.
        auto_parse_labels: Automatically extract "@label" strings from the task content.
        sync: Start by synching the Sync API cache (recommended, unless testing).
        commit: End by committing the added task to the Todoist server (recommended, unless testing).
        show_queue: Show API queue before submitting the changes to the server.
        verbose: Be extra verbose when printing information during function run.

    Returns:
        The newly added task (todoist.models.Item object).

    """
    api = get_todoist_api()
    if sync:
        api.sync()
    if verbose >= 0:
        print(f"\nAdding new task:", file=sys.stderr)
        print(f" - content:", content, file=sys.stderr)
        print(f" - project:", project, file=sys.stderr)
        print(f" - labels:", list(labels) if labels else None, file=sys.stderr)
        print(f" - priority:", priority, file=sys.stderr)
        print(f" - due:", due, file=sys.stderr)
        print(f" - note:", note, file=sys.stderr)

    params = {}

    if due:
        params['due'] = {"string": due}

    if project:
        if isinstance(project, str):
            # We need to find project_id from project_name:
            project_name = project
            try:
                project_id = next(iter(p['id'] for p in api.projects.all() if p['name'].lower() == project_name.lower()))
            except StopIteration:
                msg = f'Project name "{project_name}" was not recognized. Please create project first. '
                print(f"\n\nERROR: {msg}\n")
                print("(You can use `todoist-cli print-projects` to see a list of available projects.)\n")
                raise ValueError(msg) from None  # raise from None to not show the StopIteration exception.
        else:
            # Project should be a project_id:
            assert isinstance(project, int)
            project_id = project
        params['project_id'] = project_id

    if labels:
        if isinstance(labels, str):
            labels = [label.strip() for label in labels.split(",")]
        if any(isinstance(label, str) for label in labels):
            # Assume label-name: We need to find label_id from label_name (lower-cased):
            labels_by_name = {label['name'].lower(): label for label in api.labels.all()}
            labels = [label if isinstance(label, int) else labels_by_name[label.lower()]['id'] for label in labels]
        params['labels'] = labels

    if priority is not None:
        params['priority'] = get_proper_priority_int(priority)

    if auto_reminder is not None:
        assert isinstance(auto_reminder, bool)
        params['auto_reminder'] = auto_reminder
    if auto_parse_labels is not None:
        assert isinstance(auto_parse_labels, bool)
        params['auto_parse_labels'] = auto_parse_labels

    # Add/create new task:
    new_task = api.items.add(content, **params)

    if verbose >= 1:
        print(f"\nNew task added:", file=sys.stderr)
        print(pformat(new_task.data), file=sys.stderr)

    if note:
        # Using a temporary ID is OK.
        new_note = api.notes.add(new_task["id"], note)
        if verbose >= 1:
            print(f"\nNew note added:", file=sys.stderr)
            print(pformat(new_note.data), file=sys.stderr)

    if show_queue:
        def get_obj_data(obj):
            return obj.data
        print("\n\nAPI QUEUE:\n----------\n")
        import json
        # This is actually the best, since it can use `default=get_obj_data` to represent model objects.
        print(json.dumps(api.queue, indent=2, default=get_obj_data))

    if commit:
        if verbose >= 0:
            print(f"\nSubmitting new task to server...", file=sys.stderr)
        api.commit()
        if verbose >= 0:
            print(f" - OK!", file=sys.stderr)
    else:
        if verbose >= 0:
            print(f"\nNew task created - but not committed to server! (because `commit=False`)", file=sys.stderr)

    return new_task


def add_tasks_cli(
        tasks_file, input_format=None, due=None, project=None, labels=None, priority=None
):
    """ Add multiple tasks from text or CSV/TSV file. """
    raise NotImplementedError("Add-tasks-CLI (for adding multiple tasks from csv/text file is not yet implemented.")


@todoist_cli.command("print-projects")  # NOT click.command().
@click.option("--print-fmt", metavar="PRINT-FORMAT")
@click.option("--sort-keys", metavar="KEYS-TO-SORT-ON")
@click.option("--sort-order", type=click.Choice(["ascending", "descending", "reversed"]))
@click.option("--sync/--no-sync", default=True)
@click.option("--verbose", "-v", count=True)
def print_projects_cli(print_fmt=None, sort_keys=None, sort_order=None, sep="\n", sync=True, verbose=0):
    """ Print projects. """
    config = get_config()
    if print_fmt is None:
        print_fmt = config.get("default_project_print_fmt", DEFAULT_PROJECT_PRINT_FMT)
    if sort_keys is None:
        sort_keys = config.get("default_project_sort_keys", DEFAULT_PROJECT_SORT_KEYS)
    if sort_order is None:
        sort_order = config.get("default_project_sort_order", DEFAULT_PROJECT_SORT_ORDER)
    if isinstance(sort_keys, str):
        sort_keys = [k.strip() for k in sort_keys.split(",")]

    api = get_todoist_api()
    if sync:
        if verbose:
            print("\nSyncing data with server...", file=sys.stderr)
        api.sync()

    projects = api.projects.all()  # Returns a list copy of api.state['projects']

    if sort_keys:
        from operator import itemgetter
        keyfunc = itemgetter(*sort_keys)
        projects = sorted(projects, key=keyfunc, reverse=(sort_order!="ascending"))

    if print_fmt == "pprint":
        pprint(projects)
    elif print_fmt == "pprint-data":
        pprint([project.data for project in projects])  # Convert to list of dicts which prints better.
    else:
        for project in projects:
            fmt_kwargs = getattr(project, 'data', project)
            fmt_kwargs.setdefault('project_name', fmt_kwargs.get('name'))
            print(print_fmt.format(**fmt_kwargs), end=sep)


if __name__ == '__main__':
    # todoist_cli()  # `todoist-cli` entry point
    # add_task_cli()  # `todoist-add-task` entry point.
    print_projects_cli()  # `todoist-cli print-projects`



# Copyright 2017–2018 Rasmus Scholer Sorensen

import sys
import todoist
import requests
# date/time packages:
# import pytz
# import pendulum
# import arrow
# import maya  # Kenneth's "Date and time for Humans" package.
# import moment
# import zulu
from actionista.todoist.config import get_token
from actionista import __version__

DEFAULT_VERBOSE_PRINT_FILE = sys.stderr


def verbose_print(*msgs, verbose=0, criteria=1, file=None):
    """ A poor-man's logging function. Because setting up a logger is just too much work. """
    if file is None:
        file = DEFAULT_VERBOSE_PRINT_FILE
    if verbose > criteria:
        print(*msgs, file=file)


def get_user_agent():
    return f"Actionista-Todoist/{__version__}"


def set_session_user_agent(session: requests.Session):
    if "User-Agent" in session.headers:
        session.headers["User-Agent"] += " " + get_user_agent()


def get_todoist_api(token=None):
    """ Returns Todoist API object with token read from file or config. """
    if token is None:
        token = get_token()
    api = todoist.TodoistAPI(token=token)
    set_session_user_agent(api.session)
    return api


def sync_and_check(api, raise_on_error=True):
    """ Sync api with servers and check that the response is good. """
    res = api.sync()  # returns a dict of parsed json data
    if 'error' in res:
        msg = f"API sync error: {res['error']} (code: {res['error_code']}, tag: {res['error_tag']}, http: {res['http_code']})"
        if raise_on_error:
            raise todoist.api.SyncError(msg)
        else:
            print(f"\n{msg}\n", file=sys.stderr)
            return False
    return res




# Copyright 2019, Rasmus Sorensen <rasmusscholer@gmail.com>
"""

Configuration module for Actionista for Todoist package and CLIs/apps.




"""
import os
import yaml


# You can use the "precision" notation to truncate long strings during
# string formatting, e.g. "{field:minwidth.maxwidth}"
DEFAULT_TASK_PRINT_FMT = (
    "{project_name:15.15} "
    # "{due_date_safe_dt:%Y-%m-%d %H:%M}  "
    # "{due_date_safe:^14}  "
    "{due_date_pretty_safe:16} "  # Omits time for all-day due dates.
    "{priority_str} "
    "{checked_str} "
    "{content:.79} "
    "(due {due_string_safe}) " 
    " {labels_str}"
)
# Alternative print_fmt examples:
# print_fmt="{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M  } {content}",
# print_fmt="{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M}  {checked_str} {content}",
# print_fmt="{project_name:15} {due_date_safe_dt:%Y-%m-%d %H:%M}  {priority_str} {checked_str} {content}",
DEFAULT_TASK_SORT_KEYS = "project_name,priority_str,content".split(",")
DEFAULT_TASK_SORT_ORDER = "ascending"
DEFAULT_PROJECT_PRINT_FMT = "{name}"
DEFAULT_PROJECT_SORT_KEYS = "name".split(",")
DEFAULT_PROJECT_SORT_ORDER = "ascending"
DEFAULT_CONFIG = {
    'default_task_print_fmt': DEFAULT_TASK_PRINT_FMT,
    'default_task_sort_keys': DEFAULT_TASK_SORT_KEYS,
    'default_task_sort_order': DEFAULT_TASK_SORT_ORDER,
    'default_project_print_fmt': DEFAULT_PROJECT_PRINT_FMT,
    'default_project_sort_keys': DEFAULT_PROJECT_SORT_KEYS,
    'default_project_sort_order': DEFAULT_PROJECT_SORT_ORDER,
}
CONFIG_PATHS = [
    "~/.todoist_config.yaml"
]
TOKEN_PATHS = [
    "~/.todoist_token.txt"
]
FILEPATHS = {
    'config': CONFIG_PATHS,
    'token': TOKEN_PATHS
}


def get_config_file(name='config'):
    fn_cands = FILEPATHS[name]
    fn_cands = map(os.path.expanduser, fn_cands)
    try:
        return next(cand for cand in fn_cands if os.path.isfile(cand))
    except StopIteration:
        return None


def get_config(config_fn=None):
    if config_fn is None:
        config_fn = get_config_file()
    if config_fn is None:
        return
    with open(config_fn) as fp:
        config = yaml.safe_load(fp)
    return config


def get_config_and_filepath():
    config_fn = get_config_file()
    if config_fn is None:
        return None, None
    with open(config_fn) as fp:
        config = yaml.safe_load(fp)
    return config, config_fn


def get_token(raise_if_missing=True, config=None):
    """ Get Todoist login token.

    Will search standard token file locations (`TOKEN_PATHS`), and if no token files are found,
    will load config and return `config['token']`.

    Returns:
        str token, or None if no token was found.

    How to obtain and install your Todoist API token:
        1. Log into your todoist.com account, go to Settings → Integrations → Copy the API token.
        2. Place the token string either in a single file in one of the paths listed in `TOKEN_PATHS`,
            or put the token in the configuration file keyed under 'token'.

    """
    token = None
    config = get_config()
    if config is not None:
        token = config.get('token')
    if token:
        return token
    fn_cands = map(os.path.expanduser, TOKEN_PATHS)
    try:
        fn = next(cand for cand in fn_cands if os.path.isfile(cand))
    except StopIteration:
        pass
    else:
        with open(fn) as fp:
            token = fp.read().strip()
    if not token and raise_if_missing:
        raise ValueError(
            "Unable to find token. Please place Todoist API token either in config file (`~/.todoist_config.yaml`)"
            "or separate `~/.todoist_token.txt` file."
        )
    return token


def store_token(new_token=None):
    """ Store a new token in the config or separate token file. """
    config, config_fn = get_config_and_filepath()
    if config and config.get('token'):
        if 'token' in config:
            print("\n - Using login token from config file:", config_fn)
        current_token = config['token']
        print(f"\n- Current token is: {current_token}")
        if new_token is None:
            new_token = input("Please enter new token (or blank to abort):")
            if not new_token:
                return current_token
        print("\n - Updating config with new token:", new_token)
        config['token'] = new_token
        with open(config_fn, 'w') as fp:
            yaml.safe_dump(config, fp)
        print(" - OK, config saved with new login token.")
        return new_token
    token_file = get_config_file(name='token') or TOKEN_PATHS[0]
    print("\n - Using login token file:", token_file)
    try:
        with open(token_file) as fp:
            current_token = fp.read().strip()
    except IOError:
        current_token = None
    print(f"\n - Current token is: {current_token}")
    if new_token is None:
        new_token = input("\nPlease enter new token (enter blank or 'Ctrl+C' to abort):  ")
        if not new_token:
            return current_token
    print(f"\n - Updating token file {token_file} with new token:", new_token)
    config['token'] = new_token
    with open(token_file, 'w') as fp:
        fp.write(new_token)
    print(" - OK, token file saved with new login token.")
    return new_token


def store_default_user_config(overwrite_existing=False):
    config_fn = get_config_file() or os.path.expanduser(CONFIG_PATHS[0])
    if os.path.exists(config_fn):
        print(f"\nConfig file already exists! ({config_fn})")
        if overwrite_existing is True:
            overwrite = True
        elif overwrite_existing is False:
            overwrite = False
        elif overwrite_existing is None:
            overwrite = (input("Do you want to overwrite the current config? [Y/n]").lower() or 'y')[0] == 'y'
        else:
            raise ValueError(f"Value '{value}' of argument 'overwrite_existing' not recognized. "
                             "Must be one of True/False/None.")
        if not overwrite:
            print(" - Skipping.")
            return
    print("\nWriting default config to file:", config_fn)
    with open(config_fn, 'w') as fp:
        yaml.safe_dump(DEFAULT_CONFIG, fp)
    print(" - OK, default config saved.")
    return DEFAULT_CONFIG



# Copyright 2019, Rasmus Sorensen <rasmusscholer@gmail.com>
"""

`todoist-action-config`: Actionista for Todoist configuration CLI.


TODO: Add support for OAuth-based authorization and token retrieval.

"""

import sys
import click
import todoist

from .config import store_token, store_default_user_config, get_token
from .utils import sync_and_check

@click.command("Actionista for Todoist configuration CLI")
@click.option("--interactive/--no-interactive", default=None)
@click.option("--store-default-config/--no-store-default-config", default=None)
@click.option("--token")
@click.option("--check-token/--no-check-token", default=None)
def todoist_config_cli(interactive=None, token=None, store_default_config=None, check_token=None):
    """ Create or update Actionista for Todoist configuration files. """
    if interactive is None:
        interactive = (not token) and (not store_default_config)
    if token or interactive:
        if interactive:
            print("""\n
Setup is configuring your API login token. The API token is a small string of numbers and 
letters, and is used in place of a password for authenticating with the Todoist servers.

You can find your personal API token by selecting 'Settings' -> 'Integrations' from 
the Todoist web app, or going to: https://todoist.com/prefs/integrations
""")
        token = store_token(new_token=token)
        if check_token is None:
            check_token = (input("\nWould you like to verify that the token works? [Y/n]  ").lower() or 'y') == 'y'
        if check_token:
            print("\nVerifying API token...")
            if token != get_token():
                print("\nWARNING: The new token does not actually match the API token stored on disk!")
            api = todoist.TodoistAPI(token=token)
            print(" - Synching API...")
            try:
                res = sync_and_check(api, raise_on_error=True)
            except todoist.api.SyncError as exc:
                print(" - ERROR!")
                print(f"\n --> {exc}")
                print("\nPlease re-check the API token and re-run this configuration command.")
                sys.exit(1)
            else:
                print(" - Success - token OK!")
    if store_default_config or interactive:
        if interactive:
            print("""\n
Setup is now checking your default configuration file. The configuration file can be used 
to customize e.g. how tasks are sorted and printed by default.""")
        store_default_user_config()


if __name__ == '__main__':
    todoist_config_cli()



# from distutils.core import setup
from setuptools import setup, find_packages
import os

# Note: Make sure to use `pip -v` when pip-installing, so you can check what is being installed.

# Note: I like the "action-cli" approach so much that I am planning to use it for other "query-oriented" CLI projects.
# Thus, there will be a shared 'actionista-base' package that provides the basic "action-chain" functionalities,
# with specialized, domain-specific packages on top.

# Regarding shared namespace between multiple distribution packages (project distributions?):
# * Put `__import__('pkg_resources').declare_namespace(__name__)` in all `actionista.__init__.py` files.
# * Add `namespace_packages =["rewind"]` argument to `setuptools.setup(...)` invocation in `setup.py`.
# * c.f. https://stackoverflow.com/a/12508037/3241277
# * c.f. http://setuptools.readthedocs.io/en/latest/setuptools.html
# * c.f. https://packaging.python.org/guides/packaging-namespace-packages/
PROJECT_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file

try:
    with open(os.path.join(PROJECT_ROOT_DIR, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except IOError:
    long_description = """
Actionista Action CLI for Todoist (actionista-todoist). 

A `find`-inspired CLI for Todoist.

Add, select, print, reschedule, modify, and complete tasks in a batch-wise fashion
from the command line.

See `README.md` for usage.

"""


# Distribution build and release:
#   python setup.py sdist
#   python setup.py bdist_wheel
#   twine upload dist/*
setup(
    name='actionista-todoist',
    version='2019.9.10',  # remember to also update __init__.py
    packages=find_packages(),  # List all packages (directories) to include in the source dist.
    url='https://github.com/scholer/actionista-todoist',
    license='GNU General Public License v3 (GPLv3)',
    author='Rasmus Scholer Sorensen',
    author_email='rasmusscholer@gmail.com',
    description='Actionista Action CLI for Todoist. '
                'Add, select, print, reschedule, modify, and complete tasks in a batch-wise fashion '
                'from the command line.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Todoist', 'Tasks', 'Productivity', 'TODO', 'GTD', 'CLI'],
    entry_points={
        'console_scripts': [
            # Action CLI entry points:
            'todoist-action-cli=actionista.todoist.action_cli:action_cli',
            'actionista-todoist=actionista.todoist.action_cli:action_cli',  # New alias

            # todoist config CLI entry points:
            'todoist-action-config=actionista.todoist.config_cli:todoist_config_cli',
            'actionista-todoist-config=actionista.todoist.config_cli:todoist_config_cli',  # Alias

            # todoist_cli entry points (click-based):
            'todoist-cli=actionista.todoist.todoist_cli:todoist_cli',
            'todoist-add-task=actionista.todoist.todoist_cli:add_task_cli',

            # Entry points for adhoc_cli commands (argparse-based):
            'todoist-adhoc=actionista.todoist.adhoc_cli:main',
            'todoist-adhoc-cli=actionista.todoist.adhoc_cli:main',  # alias
            'todoist_today_or_overdue=actionista.todoist.adhoc_cli:print_today_or_overdue_tasks',
        ],
        # 'gui_scripts': [
        # ]
    },
    # pip will install these modules as requirements.
    install_requires=[
        'todoist-python',   # official Todoist python API from Doist
        'pyyaml',           # Config loading.
        'pytz',             # Timezone support.
        'python-dateutil',  # Parsing ISO timestamps.
        'dateparser',       # Required for human_date_to_iso()
        'parsedatetime',    # Has better concept of accuracy of parsed date/time than dateparser.
        'click',            # CLI package. (Only used for auxiliary CLI programs)
    ],
    python_requires='>=3.6',  # Type-hints, f-strings,
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: End Users/Desktop',
        # 'Intended Audience :: Science/Research',
        # 'Intended Audience :: Developers',
        # 'Intended Audience :: Education',
        # 'Intended Audience :: Healthcare Industry',

        # 'Topic :: Software Development :: Build Tools',
        # 'Topic :: Education',
        # 'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match 'license' above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: POSIX :: Linux',
    ],
)



# Copyright 2017–2018 Rasmus Scholer Sorensen

"""

This module contains a set of pre-defined todoist command line actions.

It is NOT a wrapper around the todoist api or any todoist library.
It is intended to be used from the comand line, or as reference examples for how to use the existing Todoist
libraries. It is not intended to be used in other code as a general Todoist library.


INSTALLATION:
-------------

1. Install this package in editable mode with pip:
    pip install -U -e .

2. Obtain a login token from the todoist website:
    Log into your todoist.com account, go to Settings → Integrations → Copy the API token.

3a. Create a file `~/.todoist_token.txt` and place your token in here, or
3b. Alternatively, create a file `~/.todoist_config.yaml` and add token as a yaml-entry:

    token: <token>



Examples, command line usage:
-----------------------------

`print-query` command:
* Note: --query defaults to "(overdue | today)"

    $ python -m actionista.todoist.adhoc_cli print-query [--query <query>] [--task-fmt <task-format>]

    $ python -m actionista.todoist.adhoc_cli print-query --print-fmt "* {due_date_time_opt_HHMM} > {title}"


`print-todays-completed-items` command:

    $ python -m actionista.todoist.adhoc_cli print-completed-today \
        --print-fmt "* {completed_date_local_dt:%H%M} > {content}" --sort-key completed_date_dt


Todoist Queries:
----------------

Refs:
* https://support.todoist.com/hc/en-us/articles/205248842-Setting-up-filters

API end point: /API/v7/query


Filter query:				What it does:
-------------				-------------
(overdue | today) & #Work	Shows all tasks that are overdue or due today that are in the “Work” project.
no date						Shows all tasks that don’t have a due date.
7 days & @waiting			Shows all tasks that are due in the next 7 days and are labelled @waiting.
created before: -365 days	Shows all tasks created more than 365 days ago.
assigned to: Allan			Shows all tasks assigned to Allan. A quick way to see what Allan’s working on.
assigned by: me				Shows all tasks you assigned to others. Handy for following up on delegated tasks.
shared & !assigned			Shows all tasks in shared project that haven’t been assigned to anyone.


Hmm... I seem to get vastly different results using API query vs web-app search/filter.
*
* However, neither seems to be correct!
* https://support.todoist.com/hc/en-us/articles/205248842-Filters



Todoist web APIs:
------------------

* Currently, the main Todoist API is the "Sync v7" API,
    available at https://todoist.com/api/v7/ (or https://api.todoist.com/API/v7/)

* Since the Sync API is a bit complicated for basic tasks, it seems Doist is also developing a basic REST API,
    available at https://developer.todoist.com/rest/v8/.
    This currently supports: Authorization and create/read/update/delete of projects/tasks/comments/labels/.



Todoist API packages:
---------------------

* https://github.com/Doist/todoist-python [installed] –
    Official python library from Doist, using requests (Session).
    Uses the Sync v7 API at 'https://todoist.com/api/v7/'.
    Has a main API object, which is synced with the server.
    Data (items/projects/notes/etc) is represented by thin model objects, which are controlled/manipulated by managers.
    Advanced functionality is delegated to the managers, e.g. ActivityManager.
    Documented at https://developer.todoist.com/.

* https://github.com/Garee/pytodoist [installed] –
    The second biggest Todoist library.
    Also uses the Sync v7 API - 'TodoistAPI.URL = https://api.todoist.com/API/v7/'
    Rolls its own API code using requests (`api` module), and object model (`todoist` module).


### About the official `python-todoist` package:

This is a little weird.
For instance, you cannot do a simple `for project in api.projects:`,
since `api.projects` is a ProjectsManager, which is not iterable.
The same goes for `api.items` and everything else that you would expect to be iterable.
You have to do `for task in api.state['items']`, or `for project in api.state['projects']`


Todoist CLIs:
----------------
* https://github.com/ddksr/cliist -
    Another Todoist CLI, rather big,
    rolls its own API using urllib requests against https://api.todoist.com/API directly.
    Also has object model, and cache system. Uses optparse for command line parsing
* https://github.com/csu/todoist-cli –
    Very simple CLI, only supports adding tasks, does not use any "libraries",
    just makes manual GET requests to https://api.todoist.com/API/addItem?content=xxx URL.
    Uses the click library for command-line interface parsing. http://click.pocoo.org/5/
* https://github.com/Matael/pydoist -
    Uses the Pydoist api library,
    acts as a simple wrapper to parse "shortened/incomplete" project names using the Levenshtein (distance) library.
* https://github.com/Raytray/todoist-cli -
    "usage: todoist.py [-h] [-p PROJECT] [-c CONTENT] [-d DUE] [-u URL] function", e.g.
    `python todoist.py query -c "tomorrow"`.
    Manually queries the API api.todoist.com/API endpoints using requests library.
* https://github.com/KryDos/todoist-cli – [last update 2014]
    Terminal-like CLI, e.g. `todoist ls tod` to list today's items.
* https://github.com/bolasblack/todoistCli - Last update 2012.
    Looks pretty horrible..
* https://github.com/akramer/NextAction -
    GTD "next action" cli interface using Todoist, rolls its API code using urllib.
* https://github.com/emorisse/todoist-tools -
    Lots of example scripts using the official todoist library.
* https://github.com/deanmalmgren/todoist-tracker -
    Tracker.. but only python2?


Notes:
--------
* There are a bunch of other Todoist libraries on github, just sort the search by "recently updated":
    https://github.com/search?l=Python&o=desc&q=todoist&s=updated&type=Repositories&utf8=%E2%9C%93
* I'm surprised there are no Todoist packages for Sublime Text yet: https://packagecontrol.io/search/todoist



References:
--------------
  (has many examples using python or curl).
* Newest API version is v7: https://todoist.com/API/v7/sync
*


Date/time libraries:
--------------------
* Python built-in `datetime` library. [so many issues: constructor and objects are intermingled, timezone hell, etc.]
* Arrow, http://arrow.readthedocs.io/en/latest/
    "Fully implemented, drop-in replacement for datetime. Timezone-aware & UTC by default."
    "Formats and parses strings automatically. Humanizes and supports a growing list of contributed locales."
    a = arrow.get(<time stamp string>)
    a.to('US/Pacific')
    a.format('YYYY-MM-DD HH:mm:ss ZZ')
    a.humanize()  # e.g. 'an hour ago'
    Does not seem to parse "an hour ago" or "tomorrow".
* Pendulum, a very complex and precise library. https://pendulum.eustace.io/
    Although, pendulum has .today(), .tomorrow(), but .parse('tomorrow') fails despite docs saying it should work.
* Maya, https://github.com/kennethreitz/maya  - By Kenneth Reitz,
    Good for adding/subtracting datetimes and timedeltas/durations.
    Uses pendulum and dateparser libraries for parsing dates.
    tomorrow = maya.when('tomorrow')  # uses dateparser.parse('tomorrow')
    time = maya.parse('01/05/2016', date_first=False)  # Uses pendulum.parse(...)
* Moment, https://github.com/zachwill/moment.
* Zulu, https://github.com/dgilland/zulu - "all objects are UTC"
* Times, https://github.com/nvie/times [deprecated, recommends Arrow instead]
* delorean, "Delorean: Time Travel Made Easy", https://github.com/myusuf3/delorean, http://delorean.rtfd.org/
* pytz, for timezones.

Parsing human natural language time strings:
* for parsing "meta dates" like "tomorrow" or "in 1 hour":
* dateutil, dateparser, parsedatetime, and datefinder, metadate
* dateutil - "Useful extensions to the standard Python datetime features."
    # https://github.com/dateutil/dateutil
    dateutil.parser.parse("tomorrow")
* parsedatetime - Parse human-readable date/time strings::
    # https://github.com/bear/parsedatetime
    cal = parsedatetime.Calendar()
    cal.parseDT("tomorrow at 6am")
* dateparser - "python parser for human readable dates" - used by Maya for parsing natural date strings.
    https://github.com/scrapinghub/dateparser
    dateparser.parse('1 hour ago', languages=['en'])
    Supports: Gregorian calendar, Persian Jalali calendar, and Hijri/Islamic Calendar.
    "Development Status :: 2 - Pre-Alpha"
* datefinder, https://github.com/akoumjian/datefinder
    Uses regexes::
        matches = datefinder.find_dates(string_with_dates)
* recurrent - "Natural language parsing of dates and recurring events."
    https://github.com/kvh/recurrent
    recurrent.RecurringEvent(now_date=datetime.datetime(2010, 1, 1)).parse('every day starting next tuesday until feb')
* Chrono - "A natural language date parser. (Python version of chrono.js)"
    # https://github.com/wanasit/chrono-python
    Seems focused on en/jp/th (english, japan, thai)
* timeparser, https://github.com/thomst/timeparser
* pytimeparse - "A small Python module to parse various kinds of time expressions."
    https://github.com/wroberts/pytimeparse
    Is for parsing natural time deltas, e.g. "1 hour", not datetimes!
* bllip-parser, https://github.com/BLLIP/bllip-parser
    "BLLIP reranking parser (also known as Charniak-Johnson parser, Charniak parser, Brown reranking parser)".


Other date/time refs:
* http://strftime.org/ - how to use datetime.strftime()
*

Library refs and surveys:
* https://docs.google.com/spreadsheets/d/1dKt0R247B8Mx5sFXd7htSOQB-B5kMODM2ydmjp9cr80/edit#gid=0



Tips & Tricks:
---------------

If you want to see how the API works, just open the web interface in Chrome,
open the "Network" tab in the Developer Tools panel, set filter to "method:POST",
and start making changes.
You can see the endpoints queried as the ":path" under "Request Headers",
and the POST data under "Form data", particularly the "commands" key.


The Todoist Sync v7 API:
-------------------------

Available at: https://todoist.com/api/v7/

Refs:
* https://developer.todoist.com/?python#items

Sync API v7 end points:
    /API/v7/sync
    /API/v7/query  (NOTE: DEPRECATED! Use sync API instead.)

    /API/v7/add_item
    /API/v7/quick/add

    /API/v7/activity/get
    /API/v7/completed
    /API/v7/completed/get_stats - Karma, etc.
    /API/v7/completed/get_all - Completed tasks.
    /API/v7/backups/get
    /API/v7/filters/get
    /API/v7/reminders/get

    templates/export_as_url
    templates/export_as_file
    templates/import_into_project

    uploads/add
    uploads/get
    uploads/delete


Requests to the /API/v7/sync endpoint:
* POST request data includes the main key 'commands', a list of dicts with commands to sync.
* A sync command is a dict (encoded as json string) with strings including
    ['type', 'args', 'uuid']

If you have a post request data, parse it as:
    d = urllib.parse.parse_qs(data)
    print(json.loads(d['commands'][0])[0].keys())


### Activity log ('activity/get') vs Completed ('completed/get_all')

Example activity log event::

    {
      "id" : 955333384,
      "object_type" : "item",
      "object_id" : 101157918,
      "event_type" : "added",
      "event_date" : "Fri 01 Jul 2016 14:24:59 +0000",
      "parent_project_id" : 174361513,
      "parent_item_id" : null,
      "initiator_id" : null,
      "extra_data" : {
        "content" : "Task1",
        "client" : "Mozilla/5.0; Todoist/830"
      }
    }


Example completed/get_all response::

    {
      "items": [
        { "content": "Item11",
          "meta_data": null,
          "user_id": 1855589,
          "task_id": 33511505,
          "note_count": 0,
          "project_id": 128501470,
          "completed_date": "Tue 17 Feb 2015 15:40:41 +0000",
          "id": 33511505
        }
      ],
      "projects": {
        # All projects with items listed above.
      }
    }



Todoist v7.1 Sync API:
----------------------

I've started preparing to upgrade to v7.1 Sync API, because I realized there was no way to
reliably determine if a task is recurring in the v7.0 API.
Unfortunately, supporting both versions adds a bit of complexity.
Specifically, the v7.1 API deals with due dates differently.
Due dates are saved as a child dict attribute, rather than just a simple string,
The format has also changed to ISO8601, and instead of using "23:59:59" as "all day" due time,
it now just sets due.date to 'YYYY-MM-DD' with no time spec.
When this is transformed to datetime objects, the time is defaulted to 0:00:00,
which would make it difficult to filter and sort due dates.
I've thus directed `inject_tasks_date_fields()` to keep the old behaviour,
setting the datetime's local time to 23:59:59.

I'm not sure how task updating work in the v7.1 Sync API.




Todoist REST API v8:
--------------------

Currently available at https://beta.todoist.com/API/v8/
* Note: This is strikingly similar to the Sync v7 API: https://todoist.com/api/v7/

> "Our original API, named Sync API, provides an easy way to deal with full and partial syncs,
> but it’s not so simple for individual calls. REST API aims to provide API developers a friendly
> way to deal with the most basic features of Todoist API."

Refs/docs:
* https://developer.todoist.com/rest/v8/


Other notes:
-------------

The Todoist web interface uses the Comet web application model:
* https://en.wikipedia.org/wiki/Comet_(programming)

Example Comet request URL:
    https://xcomet02.todoist.com/comet/1497916545644/?js_callback=CometChannel.scriptCallback&
    channel=user-111xxx28-41cxxx5e7faa8xxxxxxx&offset=-1&_client_id=_td_1497909186xxxx


"""

import datetime
from datetime import timezone
import dateutil.parser
from dateutil.parser import parse as parse_date
from pprint import pprint
import argparse

from todoist.models import Project

from actionista.todoist.tasks_utils import inject_tasks_date_fields, parse_tasks_content, inject_tasks_project_fields
from actionista.todoist.utils import get_todoist_api


def todoist_query(query, token=None):
    """ Get tasks matching the given query.

    Note: This currently uses the `query` endpoint at https://todoist.com/API/v7/query, which has been deprecated.
    Apparently the `query` API endpoint had quality issues,
    c.f. https://github.com/Doist/todoist-api/issues/22 and https://github.com/Doist/todoist-python/issues/37.
    They currently recommend "client side filtering" after obtaining all todoist data via the Sync API.
    It seems they are building a new REST API ("v8") which allows for simple task queries using a filter string.

    Args:
        query: Filter/query string to search for tasks.
        token: Login token. If not provided, will try to load token from file or config.

    Returns:
        tasks: list of task dicts

    """
    api = get_todoist_api(token)
    # If you are going to use any of the object models, you should always start by invoking `api.sync()`,
    # which fetches and caches the whole Todoist state (and also returns it).
    # If you are just going to use api.query(queries), then you don't need to sync.
    # Note that the query endpoint was deprecated due to quality issues, sometimes returning unexpected results.
    # TODO: The `qyery` endpoint on the v7 "Sync" API has been deprecated.
    # TODO: Consider using saved filters via todoist.managers.filters.FilterManager,
    # TODO: Or alternatively use the new v8 "REST" API.
    # Edit: FiltersManager is only to create/read/update/delete filters, not applying filters. Same for models.Filter.
    query_res = api.query(queries=[query])  # `api.query` expects a sequence of queries
    tasks = query_res[0]["data"]
    return tasks


def filter_tasks(tasks, filter):
    """ Filter a list of tasks. Reference function.

    Args:
        tasks: List of task dicts.
        filter: Either
            (a) A dict or list of (key, filter_value) pairs used as `task[key] == filter_value`, or
            (b) a function that returns True for all elements that passes the filter.

    Returns:
        Filtered list of tasks.

    Examples:
        >>> filter_tasks(tasks, filter={'content': "My task", "checked": 0, })
    """
    if isinstance(filter, dict):
        filter = list(filter.items())
    if isinstance(filter, list):
        print(filter)
        def filter_func(task):
            # print(filter)
            return all(task[k] == v for k, v in filter)
    else:
        filter_func = filter
    tasks = [task for task in tasks if filter_func(task)]
    return tasks


def sort_tasks(tasks, sort_key="default"):
    """ Sort a list of tasks. Reference function. Default currenltly means "Sort by due date".

    The only noteworthy code here is the default sort by due date, where we use "2100-01-01"
    instead of None (because datetime and None cannot be compared).

    Args:
        tasks: A list of task dicts.
        sort_key: Sorting key(s) or function. Either:
            (a) A single key to sort by, or
            (b) a sequence of multiple keys to sort by, or
            (c) a key function (in which case, just use `tasks.sort(key=keyfunc)` directly?).

    Returns:
        Sorted list of tasks.
    """
    if sort_key is None or sort_key == "default":
        # changed: `due_date` is now `due_date_utc`
        sort_key_func = lambda task: task["due_date_dt"] or datetime.datetime(2100, 1, 1)
    elif isinstance(sort_key, str):
        # Note: do not do `sort_key = lambda task: task[sort_key]` because the lambda function is only evaluated
        # after `sort_key` is updated (to be the lambda function) - weird result of Python's closure mechanism.
        sort_key_func = lambda task: task[sort_key]
    elif isinstance(sort_key, (list, tuple)):
        sort_key_func = lambda task: tuple(task[k] for k in sort_key)
    else:
        sort_key_func = sort_key
    tasks.sort(key=sort_key_func)
    return tasks


def process_tasks(tasks, sort_key="default", filter=None, parse_task=True, task_regex=None):
    """ Parse dates, parse content, filter, sort -
    Args:
        tasks:
        sort_key:
        filter:
        parse_task:
        task_regex:

    Returns:
        Processed list of tasks.
    """

    # Parse date strings and create datetime objects (*_dt):
    inject_tasks_date_fields(tasks)

    # Parse custom metadata from task content (using `TASK_REGEX` regular expression):
    if parse_task:
        parse_tasks_content(tasks, task_regex=task_regex)

    if filter:
        tasks = filter_tasks(tasks, filter=filter)

    if sort_key:
        tasks = sort_tasks(tasks, sort_key=sort_key)

    return tasks


def print_tasks(
        tasks, print_fmt="{content}", sep="\n",
        sort_key="default", filter=None,
        parse_task=True, task_regex=None,
    ):
    """ Parse dates, parse content, filter, sort, and finally print a list of tasks.

    Returns:
        The processed (parsed/filtered/sorted) list of tasks.
    """

    if print_fmt == 'pprint':
        pprint(tasks)
    elif print_fmt == 'repr':
        print(tasks)
    else:
        print(sep.join(print_fmt.format(**task) for task in tasks))

    return tasks


def print_query_result(
        query, token=None, print_fmt="{content}", sort_key="default", filter=None,
        parse_task=True, task_regex=None,
):
    """Perform a search using a single query string and print the results.

    Args:
        query:
        token:
        task_fmt:
        sort_key:
        task_regex: Parse task['content'] and add

    Returns:
        List of results

    Examples:
        >>> token = get_token()
        >>> res = print_query_result("@reward", print_fmt="{content} - {checked}")  # print all task with rewards

        >>> res = print_query_result(
        >>>     "@reward", print_fmt="{title}\t{reward}}",
        >>>     task_regex="^(?P<title>.*?)\s*(?P<reward_group>\{reward:\s?(?P<reward>.+)\})?\s*$")

    If you do this manually using the TodoistAPI class:
        todoist.TodoistAPI(token=token)

    """
    tasks = todoist_query(query, token)
    # Returns filtered, sorted task list
    tasks = print_tasks(tasks, print_fmt=print_fmt, sort_key=sort_key,
                        filter=filter, parse_task=parse_task, task_regex=task_regex)
    return tasks


def print_todays_tasks(
        token=None, print_fmt="{content}", sort_key="default"
):
    """Print all tasks that are due today."""
    return print_query_result(query="today", token=token, print_fmt=print_fmt, sort_key=sort_key)


def print_overdue_tasks(
        token=None, print_fmt="{content}", sort_key="default"
):
    """Print all overdue tasks."""
    return print_query_result(query="overdue", token=token, print_fmt=print_fmt, sort_key=sort_key)


def print_today_or_overdue_tasks(
        token=None, print_fmt="{content}", sort_key="default"
):
    """Print all tasks that are either overdue or due today."""
    # OBS: If you return anything from a CLI entry point, the returned value is printed to stdout.
    ret = print_query_result(query="overdue | today", token=token, print_fmt=print_fmt, sort_key=sort_key)


def get_overdue_items(token=None, query="overdue", incl_time=False):
    """ Get overdue tasks.
    Currently, this just invokes `api.sync()` and then filtering the response manually.

    It may be more appropriate to either:
        (a) Use a filter, either a saved filter (using `api.filters), or
        (b) Use the (now deprecated) `query/` endpoint, or
        (c) Use the REST API at https://beta.todoist.com/API/v8/

    """
    api = get_todoist_api(token)
    api.sync()
    now = datetime.datetime.now(timezone.utc)
    # today = datetime.fromordinal(now.toordinal())  # ordinal only includes date not time.
    today = datetime.datetime(*now.timetuple()[:3])
    cut_date = now if incl_time else today
    # items = (item for item in api.items.all() if item['due_date_utc'])
    # overdue = [item for item in items if arrow.get(parse_date(item['due_date_utc'])) < cut_date]
    # overdue = [item for item in items if parse_date(item['due_date_utc']) < cut_date]
    # overdue = [item for item in items if arrow.get(parse_date(item['due_date_utc'])) < today]  # Using arrow
    # use timestamp (seconds from epoch):
    # overdue = [item for item in items if parse_date(item['due_date_utc']).timestamp() < cut_date.timestamp()]
    overdue = [item for item in api.items.all()
               if item['due_date_utc'] is not None
               and parse_date(item['due_date_utc']).timestamp() < today.timestamp()]  # Timestamp is epoch float
    return overdue


def completed_get_all(
        token=None, verbose=1,
        project_id=None, since=None, until=None, limit=None, offset=None, annotate_notes=None,

):
    """ Get completed items, using the 'completed/get_all' endpoint (via CompletedManager).

    This endpoint is more specific than the Activity Log endpoint.
    It still returns a dict with "items" and "projects",
    but the data keys are more descriptive and specific (e.g. "completed_date" instead of "event_date").

    Note: This API endpoint is only available for Todoist Premium users.

    Refs/docs:
    * https://developer.todoist.com/?python#get-all-completed-items

    Available filter parameters/keys/kwargs for the 'completed/get_all' endpoint includes:
    * project_id: Only return completed items under a given project.
    * since, until: return event after/before these time (formatted as 2016-06-28T12:00).
    * limit, offset: Maximum number of events to return, and paging offset.
    * annotate_notes: Return notes together with the completed items (a true or false value).

    Note that while the regular Sync API uses ridiculous time format, the 'completed' API takes sane
    ISO-8601 formatted datetime strings for `since` and `until` params.

    Returns:
        A two-tuple of (items, projects).
        * Although maybe it is better to just return the response dict
            which contains two keys, "items" and "projects"?

    Note:
        * I'm having issues with the project_id for tasks not matching any projects.
        * The projects returned are good; it is indeed an issue with the project_id value. Maybe it is an old value?
        * This issue doesn't seem to be present for the `activity/get` endpoint.
        * This issue is NOT described in the docs, https://developer.todoist.com/sync/v7/#get-all-completed-items
        * The documented example makes it appear that item['project_id'] should match projects[id]['id'] .

    """
    api = get_todoist_api(token)
    kwargs = dict(
        project_id=project_id,
        since=since,
        until=until,
        limit=limit,
        offset=offset,
        annotate_notes=annotate_notes,
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if verbose:
        print("\nRetrieving completed tasks from the Todoist `completed/get_all` endpoint...")
    res = api.completed.get_all(**kwargs)
    # from pprint import pprint; pprint(res)
    return res['items'], res['projects']


def activity(
        token=None,
        object_type=None, object_id=None, event_type=None, object_event_types=None,
        parent_project_id=None, parent_item_id=None, initiator_id=None,
        since=None, until=None,
        limit=None, offset=None,
):
    """ Search using the activity log ('activity/get') endpoint.

    This endpoint can be used to retrieve all recent activity, all activity for a given project, etc.

    Refs/docs:
    * https://developer.todoist.com/?python#get-activity-logs

    Note: api.activity.get() without parameters will get all activity (limited to 30 entries by default).

    Endpoint parameters/keys are:
    * object_type (str): e.g. 'item', 'note', 'project', etc.
    * object_id (int): Only show for a particular object, but only if `object_type` is given.
    * event_type (str): e.g. 'added', 'updated', 'completed', 'deleted',
    * object_event_types: Combination of object_type and event_type.
    * parent_project_id: Only show events for items or notes under a given project.
    * parent_item_id: Only show event for notes under a given item.
    * initiator_id: ?
    * since, until: return event after/before these time (formatted as 2016-06-28T12:00).
    * limit, offset: Maximum number of events to return, and paging offset.

    Returns:
        events: A list of event dicts.

    """
    api = get_todoist_api(token)
    kwargs = dict(
        object_type=object_type, object_id=object_id, event_type=event_type, object_event_types=object_event_types,
        parent_project_id=parent_project_id, parent_item_id=parent_item_id, initiator_id=initiator_id,
        since=since, until=until,
        limit=limit, offset=offset,
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    events = api.activity.get(**kwargs)
    return events


def get_todays_completed_events(token=None, sync=True, verbose=1):
    """ Get all today's completed events using ActivityManager (against the 'activity/get' endpoint).
    See `activity()` for info.
    Since only tasks can be completed, this is basically just another way to get completed tasks.
    Better alternative: Use `get_todays_completed_items()`, which uses the CompletedManager.
    """
    api = get_todoist_api(token)
    if sync:
        if verbose:
            print("Syncing with the server...")
        api.sync()
    kwargs = dict(
        since="{:%Y-%m-%dT06:00}".format(datetime.date.today()),
        object_type='item',
        event_type='completed',
        limit=40,
    )
    events = api.activity.get(**kwargs)
    return events


def get_todays_completed_items(token=None, verbose=1):
    """ Get todays completed items, using CompletedManager against the 'completed/get_all' endpoint.
    See `completed_get_all()` for info.
    Note: The completed/get_all endpoint seems to return tasks with wrong project_id values.
    Perhaps use activity/get endpoint via get_todays_completed_events() instead?
    """
    # TODO: Specifying since 'today' this way is bit of a hack; only works for timezones near or west of UTC.
    # TODO: Make proper, generic way of dealing with time and time strings!
    since="{:%Y-%m-%dT06:00}".format(datetime.date.today())
    if verbose and verbose >= 2:
        print("Returning all completed tasks since UTC:", since)
    return completed_get_all(token=token, verbose=verbose, since=since)


def print_todays_completed_items(
        token=None, print_fmt="{content}", sort_key="default",
        add_project_info=True,
):
    """ Print all tasks that were completed today.

    There are at least two ways to get "completed tasks":

        (a) Sync all tasks as normal and filter those that have a `completed_date` field equalling today.
        (b) Use the `completed/get_all` endpoint.
        (c) Use the `activity` endpoint to get 'events'.

    Regarding (a), you could just use the actionista todoist-action-cli, except that the
        `date_completed` field is always None for tasks returned by the Todoist v7.0 Sync API.

    If using (b), then there seem to be an issue with some tasks having the wrong project id.
    This can, however, be mitigated by passing `strict=False` to `inject_tasks_project_fields()`.
    The `completed/get_all` endpoint returns completed items (list) and their related projects (dict) as follows:
        {
          "items": [
            { "content": "Item11", "meta_data": null, "user_id": 1855589, "task_id": 33511505, "note_count": 0,
              "project_id": 128501470,
              "completed_date": "Tue 17 Feb 2015 15:40:41 +0000",
              "id": 33511505
            }
          ],
          "projects": {
            "128501470":
            { "color": 7, "collapsed": 0, "indent": 1, "is_deleted": 0,
              "id": 128501470, "user_id": 1855589,
              "name": "Project1", "item_order": 36, "is_archived": 0 }
          }
        }

    If using (c), then be aware that event data is structured differently than normal tasks:
    Example events list with one event:

        [{'event_date': 'Wed 02 May 2018 08:27:51 +0000',
          'event_type': 'completed',
          'extra_data': {'client': 'Mozilla/5.0; Todoist/901',
                         'content': 'Pick up USB cable'},
          'id': 4035856215,
          'initiator_id': None,
          'object_id': 2631967385,
          'object_type': 'item',
          'parent_item_id': None,
          'parent_project_id': 192280257}]

    """
    # events = get_todays_completed_events(token=token)  # Use activity/get instead of completed/get_all.
    # print("Events:")
    # pprint(events)
    # return
    # The 'activity' endpoint is generic: instead of returning task items, it returns generic events.
    # The task name/content is available as event['extra_data']['content'], and
    # the task project_id is available as event['parent_project_id'].
    # So we cannot use the event directly; we would have to either transform them,
    # or use a special printer and print_fmt to print the events.
    tasks, projects = get_todays_completed_items(token=token)  # currently gives tasks with erroneous project_id.
    # from pprint import pprint; pprint(tasks); print(); pprint(projects)
    if add_project_info:
        # NOTICE: the project_id and projects returned by api.completed.get_all()
        # don't always match up. The task's project_id is sometimes wrong!
        # http://todoist.com/API/v7/completed/get_all
        # Maybe they are just using old project_id values, but still.
        # Is the api.activity.get() endpoint at https://todoist.com/API/v7/activity/get any better?
        # Yes, it seems this gives correct project_id values for the returned tasks.
        # Edit: I've added a non-strict option to inject_tasks_project_fields, so we can still use completed/get_all data.
        inject_tasks_project_fields(tasks, projects, strict=False)
    print("\nToday's completed tasks:\n")
    tasks = print_tasks(tasks, print_fmt=print_fmt, sort_key=sort_key)
    return tasks


def reschedule_items(items, new_date='today'):
    """

    Args:
        items:

    Returns:

    How should task rescheduling be done?

    Rescheduling recurring tasks:
        Update `due_date_utc`, changing the date part to today but keeping the original time part.
        Keep `date_string` as-is.

    Rescheduling non-recurring tasks:
        Update `due_date_utc`, changing the date part to today but keeping the original time part.
        Update `date_string` to "%m %d %H:%M" if date_string has a time, otherwise just "%m %d".
        Note: If due_date_utc does not have a time but is date-only, then the time will be 23:59:59, not 00:00.
        Which makes sense: task is due by the end of the day. OBS: The 59 seconds matter, 23:59:58 doesn't work.

    Example of request command to reschedule a task with id=2233091342 to Thur 22 Jun 2017 14:00:00 (EST; 18:00:00 UTC)
        [{
            "type":"item_update",
            "args":{"id":2233091342,"due_date_utc":"2017-6-22T18:00:00"},
            "uuid":"c51c443e-1316-9830-fa67-57aafd2815b1"
        }]
    (uuid is used to prevent executing duplicate or old requests)

    We can use ItemManager.update_date_complete() which uses the 'item_update_date_complete' command type:
        [{
            'type': 'item_update_date_complete',
            'uuid': 'a4663f86-5539-11e7-9273-38c9863570b9',
            'args': {'id': 2233091342, 'new_date_utc': '2017-06-21T00:00', 'date_string': 'every day at 10 pm'}
        }]
    Note: According to the docs this is mostly to mark recurring tasks as complete.

    If `new_date_utc` is wrong, you will get a SyncError:
        SyncError: (
            '87b2186e-5538-11e7-b699-38c9863570b9', {
                'error_extra': {'expected': 'date: YYYY-MM-DDTHH:MM', 'argument': 'new_date_utc'},
                'error_tag': 'INVALID_ARGUMENT_VALUE', 'error': 'Invalid argument value',
                'command_type': 'item_update_date_complete', 'error_code': 20, 'http_code': 400})
    """
    # Expected date format: YYYY-MM-DDTHH:MM  (in UTC time!)
    if isinstance(new_date, datetime.datetime):
        new_date = new_date.strftime()
    else:
        try:
            datetime.datetime.strptime(new_date, "YYYY-MM-DDTHH:MM")
        except ValueError:
            # Assume new_date is in some kind of "natural language" format.
            import pendulum
            import maya
            import arrow
            import dateutil
            new_date = maya.parse(new_date)
            # Make sure the new due date is in UTC!
            new_date = "{:%y-%m-%dT%H:%M".format(new_date)
    # all_day_time = datetime.time
    # import maya
    for item in items:
        item['due_date_utc'] = new_date


def reschedule_cmd(query, new_date='today'):
    """

    Args:
        query:
        new_date:

    Returns:


    Discussion:


    Example POST request form data using the web interface to "reschedule" a single item:
        resource_types:["all"]
        commands: [{
            "type": "item_update",
            "args": {"date_string":"Aug 7", "date_lang":"en", "id":2235317410, "due_date_utc": "2017-8-8T03:59:59"},
            "uuid":"9b5ba425-7dcf-db4d-5a2a-d870bf731xxx"
        }]
        day_orders_timestamp:1502140199.43
        collaborators_timestamp:
        sync_token:xxxx
        with_web_static_version:true
        limit_notes:1
        max_notes:5
        with_dateist_version:1
        _client_id:_td_1501792xxxxx

    OBS: To see how the web interface uses the sync api, just open the "Network" tab in Chrome's developer tools
    and look for form data for requests to the /API/v7/sync endpoint.

    """
    # Q: Do we need an actual ISO date-string, or can we use the Todoist API to parse `new_date`?
    # Q: If we need an ISO date-string, can we use e.g. maya/arrow/pendulum to parse `new_date`?
    # Q: Which is better, to use the full "sync" API, or to reschedule individual tasks?
    # Q:
    # tasks = todoist_query(query)
    api = get_todoist_api()
    query_res = api.query(queries=[query])  # sequence of queries
    tasks = query_res[0]["data"]
    reschedule_items(tasks, new_date=new_date)
    api.sync()
    # Workflow:
    # 1. Sync:
    #   api.sync() -> api._update_state() -> populates api.state['items']
    # 2. Update:
    #   api.items.update(item_id, **kwargs) -> api.queue.append(cmd)
    # 3. Commit:
    #   api.commit() -> api.sync(commands=api.queue) -> api._post()

    date_string, due_date_utc, datetime_obj = dateutil.parser.parse(new_date)

    # api.items.all() is same as api.state['items']
    for item in api.items.all():
        # Yes, these are actual `model.Item` objects. Made during `api._update_state()` with::
        #    newobj = model(remoteobj, self); self.state[datatype].append(newobj)  # self = api
        item.update(date_string=date_string, due_date_utc=due_date_utc)

    api.commit()


def print_projects(print_fmt="pprint-data", sort_key=None, sync=True, sep="\n"):

    api = get_todoist_api()
    if sync:
        print("\nSyncing data with server...")
        api.sync()
    projects = api.state['projects']  # Not api.projects, which is a non-iterable ProjectsManager. Sigh.
    if print_fmt == "pprint":
        pprint(projects)
    elif print_fmt == "pprint-data":
        pprint([project.data for project in projects])  # Convert to list of dicts which prints better.
    else:
        for project in projects:
            fmt_kwargs = project.data if isinstance(project, Project) else project
            print(print_fmt.format(**fmt_kwargs), end=sep)


def parse_args(argv=None):
    ap = argparse.ArgumentParser(prog="RS Todoist python CLI")
    subparsers = ap.add_subparsers(dest='command')
    # There are two ways to describe which subcommand to invoke depending on the subparser argument passed by the user.
    #  a. Capturing the subparser argument explicitly using the `dest` argument in ap.add_subparsers().
    #  b. Let the sub-parser set the subcommand function using sp.set_defaults(func=foo).

    # TODO: Use the method from `pptx-downsizer` to get default arguments from functions. (Or, use click.)
    sp = subparsers.add_parser('print-query')
    sp.set_defaults(func=print_query_result)
    sp.add_argument('--query', default="(overdue | today)")  # "(overdue | today) & #Work"

    # print_todays_completed_items
    sp = subparsers.add_parser('print-completed-today')
    sp.set_defaults(func=print_todays_completed_items)

    # print_todays_completed_items
    sp = subparsers.add_parser('print-today-or-overdue')
    sp.set_defaults(func=print_today_or_overdue_tasks)

    # Add shared args for task commands:
    for cmd, sp in subparsers.choices.items():
        sp.add_argument('--print-fmt', default="{content}")
        sp.add_argument('--sort-key', default="default")

    # print_projects
    sp = subparsers.add_parser('print-projects')
    sp.set_defaults(func=print_projects)
    sp.add_argument('--print-fmt', default="pprint")
    sp.add_argument('--sort-key', default="default")

    # OBS: Do not make special switches, just use print-fmt and the regex-parsed variables.
    argns = ap.parse_args(argv)

    return argns


def main(argv=None):
    """ Main CLI entry point.

    Args:
        argv: Command line arguments (default: sys.argv)

    Returns:
        None

    Examples:
        $ todoist print-today-or-overdue --print-fmt "* {title}"
        $ todoist print-completed-today --print-fmt "* {title}"

    For CLI usage, type:

        $ todoist-achoc --help

    """

    argns = parse_args(argv)
    args = vars(argns)
    command, func = args.pop('command'), args.pop('func', None)  # command = subparser name; func = function obj.
    # If action-cli is invoked without a command argument, then func is None:
    if func is None:
        print(main.__doc__)
        return
    res = func(**args)
    # print_todays_tasks()
    # print_query_result(query="(overdue | today) & #Work")  # Note: 'query' endpoint deprecated because weird results.
    # print_query_result(query="no date")


if __name__ == '__main__':
    main()
